from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

import narwhals as nw

if TYPE_CHECKING:
    from narwhals.typing import FrameT


def query(
    customer_ds: FrameT, nation_ds: FrameT, lineitem_ds: FrameT, orders_ds: FrameT
) -> FrameT:
    var1 = datetime(1993, 10, 1)
    var2 = datetime(1994, 1, 1)

    return (
        customer_ds.join(orders_ds, left_on="c_custkey", right_on="o_custkey")
        .join(lineitem_ds, left_on="o_orderkey", right_on="l_orderkey")
        .join(nation_ds, left_on="c_nationkey", right_on="n_nationkey")
        .filter(nw.col("o_orderdate").is_between(var1, var2, closed="left"))
        .filter(nw.col("l_returnflag") == "R")
        .with_columns(
            (nw.col("l_extendedprice") * (1 - nw.col("l_discount"))).alias("revenue")
        )
        .group_by(
            "c_custkey",
            "c_name",
            "c_acctbal",
            "c_phone",
            "n_name",
            "c_address",
            "c_comment",
        )
        .agg(nw.sum("revenue"))
        .select(
            "c_custkey",
            "c_name",
            "revenue",
            "c_acctbal",
            "n_name",
            "c_address",
            "c_phone",
            "c_comment",
        )
        .sort(by="revenue", descending=True)
        .head(20)
    )



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

import pytest

import narwhals as nw
from tests.utils import PANDAS_VERSION, POLARS_VERSION, assert_equal_data, is_windows

if TYPE_CHECKING:
    from tests.utils import Constructor, ConstructorEager

data = {"a": [[3, None, 2, 2, 4, None], [-1], None, [None, None, None], [], [3, 4, None]]}
expected = [2.5, -1, None, None, None, 3.5]
expected_pyarrow = [2.5, -1, None, None, None, 3]


def test_median_expr(request: pytest.FixtureRequest, constructor: Constructor) -> None:
    if any(backend in str(constructor) for backend in ("dask", "cudf", "ibis")) or (
        "polars" in str(constructor) and POLARS_VERSION < (0, 20, 7)
    ):
        # ibis issue: https://github.com/ibis-project/ibis/issues/11788
        request.applymarker(pytest.mark.xfail)
    if "pandas" in str(constructor):
        if PANDAS_VERSION < (2, 2):
            pytest.skip()
        pytest.importorskip("pyarrow")
    if (
        any(backend in str(constructor) for backend in ("pandas", "pyarrow"))
        and sys.version_info < (3, 10)
        and is_windows()
    ):  # pragma: no cover
        reason = "The issue only affects old Python versions on Windows."
        request.applymarker(pytest.mark.xfail(reason=reason))
    result = nw.from_native(constructor(data)).select(
        nw.col("a").cast(nw.List(nw.Int32())).list.median()
    )
    if any(
        backend in str(constructor)
        for backend in ("pandas", "pyarrow", "pandas[pyarrow]")
    ):
        # there is a mismatch as pyarrow uses an approximate median
        assert_equal_data(result, {"a": expected_pyarrow})
    else:
        assert_equal_data(result, {"a": expected})


def test_median_series(
    request: pytest.FixtureRequest, constructor_eager: ConstructorEager
) -> None:
    if any(backend in str(constructor_eager) for backend in ("cudf",)) or (
        "polars" in str(constructor_eager) and POLARS_VERSION < (0, 20, 7)
    ):
        request.applymarker(pytest.mark.xfail)
    if "pandas" in str(constructor_eager):
        if PANDAS_VERSION < (2, 2):
            pytest.skip()
        pytest.importorskip("pyarrow")
    if (
        any(backend in str(constructor_eager) for backend in ("pandas", "pyarrow"))
        and sys.version_info < (3, 10)
        and is_windows()
    ):  # pragma: no cover
        reason = "The issue only affects old Python versions on Windows."
        request.applymarker(pytest.mark.xfail(reason=reason))
    df = nw.from_native(constructor_eager(data), eager_only=True)
    result = df["a"].cast(nw.List(nw.Int32())).list.median()
    if any(
        backend in str(constructor_eager)
        for backend in ("pandas", "pyarrow", "pandas[pyarrow]")
    ):
        # there is a mismatch as pyarrow uses an approximate median
        assert_equal_data({"a": result}, {"a": expected_pyarrow})
    else:
        assert_equal_data({"a": result}, {"a": expected})



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import narwhals as nw

pytest.importorskip("pyarrow")
import pyarrow as pa
import pyarrow.compute as pc

if TYPE_CHECKING:
    from tests.utils import ConstructorEager


def test_to_arrow(constructor_eager: ConstructorEager) -> None:
    data = [1, 2, 3]
    result = nw.from_native(constructor_eager({"a": data}), eager_only=True)[
        "a"
    ].to_arrow()

    assert pa.types.is_int64(result.type)
    assert pc.all(pc.equal(result, pa.array(data, type=pa.int64())))


def test_to_arrow_with_nulls(
    constructor_eager: ConstructorEager, request: pytest.FixtureRequest
) -> None:
    if "pandas_constructor" in str(constructor_eager) or "modin_constructor" in str(
        constructor_eager
    ):
        request.applymarker(pytest.mark.xfail)

    data = [1, 2, None]
    result = (
        nw.from_native(constructor_eager({"a": data}), eager_only=True)["a"]
        .cast(nw.Int64)
        .to_arrow()
    )

    assert pa.types.is_int64(result.type)
    assert pc.all(pc.equal(result, pa.array(data, type=pa.int64())))



from __future__ import annotations

import pytest

import narwhals as nw
from tests.utils import (
    PANDAS_VERSION,
    POLARS_VERSION,
    Constructor,
    ConstructorEager,
    assert_equal_data,
    uses_pyarrow_backend,
)

data = {"a": ["-1", "+1", "1", "12", "123", "99999", "+9999", None]}
expected = {"a": ["-01", "+01", "001", "012", "123", "99999", "+9999", None]}


def test_str_zfill(request: pytest.FixtureRequest, constructor: Constructor) -> None:
    if uses_pyarrow_backend(constructor) and PANDAS_VERSION < (3,):
        reason = (
            "pandas with pyarrow backend doesn't support str.zfill, see "
            "https://github.com/pandas-dev/pandas/issues/61485"
        )
        request.applymarker(pytest.mark.xfail(reason=reason))

    if "pandas" in str(constructor) and PANDAS_VERSION < (1, 5):
        reason = "different zfill behavior"
        pytest.skip(reason=reason)

    if "polars" in str(constructor) and POLARS_VERSION < (0, 20, 5):
        reason = (
            "`TypeError: argument 'length': 'Expr' object cannot be interpreted as an integer`"
            "in `expr.str.slice(1, length)`"
        )
        pytest.skip(reason=reason)

    df = nw.from_native(constructor(data))
    result = df.select(nw.col("a").str.zfill(3))
    assert_equal_data(result, expected)


def test_str_zfill_series(
    request: pytest.FixtureRequest, constructor_eager: ConstructorEager
) -> None:
    if uses_pyarrow_backend(constructor_eager) and PANDAS_VERSION < (3,):
        reason = (
            "pandas with pyarrow backend doesn't support str.zfill, see "
            "https://github.com/pandas-dev/pandas/issues/61485"
        )
        request.applymarker(pytest.mark.xfail(reason=reason))

    if "pandas" in str(constructor_eager) and PANDAS_VERSION < (1, 5):
        reason = "different zfill behavior"
        pytest.skip(reason=reason)

    if "polars" in str(constructor_eager) and POLARS_VERSION < (0, 20, 5):
        reason = (
            "`TypeError: argument 'length': 'Expr' object cannot be interpreted as an integer`"
            "in `expr.str.slice(1, length)`"
        )
        pytest.skip(reason=reason)

    df = nw.from_native(constructor_eager(data), eager_only=True)
    result = df["a"].str.zfill(3)
    assert_equal_data({"a": result}, expected)



# ruff: noqa: ARG003
from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest

pytest.importorskip("pandas")


import pandas as pd

import narwhals as nw

if TYPE_CHECKING:
    from collections.abc import Sequence

    from typing_extensions import Self


class CustomInt16Dtype(pd.api.extensions.ExtensionDtype):  # pragma: no cover
    name = "custom_int_16"
    type = int
    _metadata = ("name",)

    @classmethod
    def construct_array_type(cls) -> type[pd.api.extensions.ExtensionArray]:  # type: ignore[valid-type]
        return CustomInt16Array


class CustomInt16Array(pd.api.extensions.ExtensionArray):  # pragma: no cover
    def __init__(self, values: Sequence[Any]) -> None:
        self._values = pd.array(values, dtype="int16")

    @classmethod
    def _from_sequence(
        cls, scalars: Sequence[Any], *, dtype: Any = None, copy: bool = False
    ) -> Self:
        return cls(scalars)

    def __len__(self) -> int:
        return len(self._values)

    @property
    def dtype(self) -> pd.api.extensions.ExtensionDtype:
        return CustomInt16Dtype()

    def copy(self) -> Self:
        return type(self)(list(self._values.copy()))


class CustomInt32Dtype(pd.api.extensions.ExtensionDtype):  # pragma: no cover
    name = "custom_int_32"
    type = int
    _metadata = ("name",)

    @classmethod
    def construct_array_type(cls) -> type[pd.api.extensions.ExtensionArray]:  # type: ignore[valid-type]
        return CustomInt32Array

    def __hash__(self) -> int:
        return hash(self.name)


class CustomInt32Array(pd.api.extensions.ExtensionArray):  # pragma: no cover
    def __init__(self, values: Sequence[Any]) -> None:
        self._values = pd.array(values, dtype="int32")

    @classmethod
    def _from_sequence(
        cls, scalars: Sequence[Any], *, dtype: Any = None, copy: bool = False
    ) -> Self:
        return cls(scalars)

    def __len__(self) -> int:
        return len(self._values)

    @property
    def dtype(self) -> pd.api.extensions.ExtensionDtype:
        return CustomInt16Dtype()

    def copy(self) -> Self:
        return type(self)(list(self._values.copy()))


def test_dataframe_with_ext() -> None:
    int16_array = CustomInt16Array([1, 2])
    int32_array = CustomInt32Array([1, 2])

    df = pd.DataFrame({"non-hash-int16": int16_array, "hash-int-32": int32_array})

    assert nw.from_native(df).schema == {
        "non-hash-int16": nw.Unknown(),
        "hash-int-32": nw.Unknown(),
    }


def test_schema_with_ext() -> None:
    pd_schema = {"non-hash-int16": CustomInt16Dtype(), "hash-int-32": CustomInt32Dtype()}
    nw_schema = nw.Schema.from_pandas_like(pd_schema)
    assert nw_schema == nw.Schema(
        {"non-hash-int16": nw.Unknown(), "hash-int-32": nw.Unknown()}
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest

import narwhals.stable.v1 as nw_v1

pytest.importorskip("polars")
import polars as pl

if TYPE_CHECKING:
    from collections.abc import Mapping

data: Mapping[str, Any] = {"a": [1, 2, 3], "b": [4.5, 6.7, 8.9], "z": ["x", "y", "w"]}


def test_interchange() -> None:
    df_pl = pl.DataFrame(data)
    df = nw_v1.from_native(df_pl.__dataframe__(), eager_or_interchange_only=True)
    series = df["a"]

    with pytest.raises(
        NotImplementedError,
        match="Cannot access native namespace for interchange-level dataframes with unknown backend",
    ):
        df.__native_namespace__()

    with pytest.raises(
        NotImplementedError,
        match="Cannot access native namespace for interchange-level series with unknown backend",
    ):
        series.__native_namespace__()


@pytest.mark.filterwarnings("ignore:.*The `ArrowDtype` class is not available in pandas")
def test_ibis(
    tmpdir: pytest.TempdirFactory, request: pytest.FixtureRequest
) -> None:  # pragma: no cover
    pytest.importorskip("ibis")
    import ibis

    try:
        ibis.set_backend("duckdb")
    except ImportError:
        request.applymarker(pytest.mark.xfail)
    df_pl = pl.DataFrame(data)

    filepath = str(tmpdir / "file.parquet")  # type: ignore[operator]
    df_pl.write_parquet(filepath)
    tbl = ibis.read_parquet(filepath)
    df = nw_v1.from_native(tbl, eager_or_interchange_only=True)
    series = df["a"]

    assert df.__native_namespace__() == ibis
    assert series.__native_namespace__() == ibis


def test_duckdb() -> None:
    pytest.importorskip("duckdb")
    import duckdb

    df_pl = pl.DataFrame(data)  # noqa: F841

    rel = duckdb.sql("select * from df_pl")
    df = nw_v1.from_native(rel, eager_or_interchange_only=True)
    series = df["a"]

    assert df.__native_namespace__() == duckdb
    assert series.__native_namespace__() == duckdb



from __future__ import annotations

from typing import TYPE_CHECKING

import narwhals as nw

if TYPE_CHECKING:
    from narwhals.typing import FrameT


def query(nation_ds: FrameT, partsupp_ds: FrameT, supplier_ds: FrameT) -> FrameT:
    var1 = "GERMANY"
    var2 = 0.0001

    q1 = (
        partsupp_ds.join(supplier_ds, left_on="ps_suppkey", right_on="s_suppkey")
        .join(nation_ds, left_on="s_nationkey", right_on="n_nationkey")
        .filter(nw.col("n_name") == var1)
    )
    q2 = q1.select(
        (nw.col("ps_supplycost") * nw.col("ps_availqty")).sum().round(2).alias("tmp")
        * var2
    )

    return (
        q1.with_columns((nw.col("ps_supplycost") * nw.col("ps_availqty")).alias("value"))
        .group_by("ps_partkey")
        .agg(nw.sum("value"))
        .join(q2, how="cross")  # pyright: ignore[reportArgumentType]
        .filter(nw.col("value") > nw.col("tmp"))
        .select("ps_partkey", "value")
        .sort("value", descending=True)
    )



from __future__ import annotations

import re

import pytest

import narwhals as nw
from tests.utils import PANDAS_VERSION, Constructor, assert_equal_data

data = {
    "OriginCityName": [
        "Columbus, OH",
        "Newark, NJ",
        "Philadelphia, PA",
        "Nashville, TN",
        "Washington, DC",
        "Phoenix, AZ",
    ]
}

expected = {
    "OriginCityName": [
        "Columbus",
        "Newark",
        "Philadelphia",
        "Nashville",
        "Washington",
        "Phoenix",
    ]
}


def test_split_list_get(request: pytest.FixtureRequest, constructor: Constructor) -> None:
    if any(backend in str(constructor) for backend in ("dask",)):
        request.applymarker(pytest.mark.xfail)

    if "pandas" in str(constructor):
        if PANDAS_VERSION < (2, 2):
            pytest.skip()
        pytest.importorskip("pyarrow")
    if (
        constructor.__name__.startswith("pandas")
        and "pyarrow" not in constructor.__name__
    ):
        df = nw.from_native(constructor(data))
        msg = re.escape("This operation requires a pyarrow-backed series. ")
        with pytest.raises(TypeError, match=msg):
            df.select(nw.col("OriginCityName").str.split(",").list.get(0))
        return
    df = nw.from_native(constructor(data))
    result = df.select(nw.col("OriginCityName").str.split(",").list.get(0))
    assert_equal_data(result, expected)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, NoReturn

from narwhals._expression_parsing import ExprKind, ExprNode
from narwhals._utils import flatten
from narwhals.expr import Expr

if TYPE_CHECKING:
    from collections.abc import Iterable
    from datetime import timezone

    from narwhals.dtypes import DType
    from narwhals.typing import TimeUnit


class Selector(Expr):
    def _to_expr(self) -> Expr:
        return Expr(*self._nodes)

    def __add__(self, other: Any) -> Expr:  # type: ignore[override]
        if isinstance(other, Selector):
            msg = "unsupported operand type(s) for op: ('Selector' + 'Selector')"
            raise TypeError(msg)
        return self._to_expr()._append_node(
            ExprNode(ExprKind.ELEMENTWISE, "__add__", exprs=(other,), str_as_lit=True)
        )

    def __or__(self, other: Any) -> Expr:  # type: ignore[override]
        if isinstance(other, Selector):
            return self._append_node(
                ExprNode(
                    ExprKind.ELEMENTWISE,
                    "__or__",
                    exprs=(other,),
                    str_as_lit=True,
                    allow_multi_output=True,
                )
            )
        return self._to_expr()._append_node(
            ExprNode(ExprKind.ELEMENTWISE, "__or__", exprs=(other,), str_as_lit=True)
        )

    def __and__(self, other: Any) -> Expr:  # type: ignore[override]
        if isinstance(other, Selector):
            return self._append_node(
                ExprNode(
                    ExprKind.ELEMENTWISE,
                    "__and__",
                    exprs=(other,),
                    str_as_lit=True,
                    allow_multi_output=True,
                )
            )
        return self._to_expr()._append_node(
            ExprNode(ExprKind.ELEMENTWISE, "__and__", exprs=(other,), str_as_lit=True)
        )

    def __rsub__(self, other: Any) -> NoReturn:
        raise NotImplementedError

    def __rand__(self, other: Any) -> NoReturn:
        raise NotImplementedError

    def __ror__(self, other: Any) -> NoReturn:
        raise NotImplementedError


def by_dtype(*dtypes: DType | type[DType] | Iterable[DType | type[DType]]) -> Selector:
    """Select columns based on their dtype.

    Arguments:
        dtypes: one or data types to select

    Examples:
        >>> import pyarrow as pa
        >>> import narwhals as nw
        >>> import narwhals.selectors as ncs
        >>> df_native = pa.table({"a": [1, 2], "b": ["x", "y"], "c": [4.1, 2.3]})
        >>> df = nw.from_native(df_native)

        Let's select int64 and float64  dtypes and multiply each value by 2:

        >>> df.select(ncs.by_dtype(nw.Int64, nw.Float64) * 2).to_native()
        pyarrow.Table
        a: int64
        c: double
        ----
        a: [[2,4]]
        c: [[8.2,4.6]]
    """
    flattened = flatten(dtypes)
    return Selector(ExprNode(ExprKind.SELECTOR, "selectors.by_dtype", dtypes=flattened))


def matches(pattern: str) -> Selector:
    """Select all columns that match the given regex pattern.

    Arguments:
        pattern: A valid regular expression pattern.

    Examples:
        >>> import pandas as pd
        >>> import narwhals as nw
        >>> import narwhals.selectors as ncs
        >>> df_native = pd.DataFrame(
        ...     {"bar": [123, 456], "baz": [2.0, 5.5], "zap": [0, 1]}
        ... )
        >>> df = nw.from_native(df_native)

        Let's select column names containing an 'a', preceded by a character that is not 'z':

        >>> df.select(ncs.matches("[^z]a")).to_native()
           bar  baz
        0  123  2.0
        1  456  5.5
    """
    return Selector(ExprNode(ExprKind.SELECTOR, "selectors.matches", pattern=pattern))


def numeric() -> Selector:
    """Select numeric columns.

    Examples:
        >>> import polars as pl
        >>> import narwhals as nw
        >>> import narwhals.selectors as ncs
        >>> df_native = pl.DataFrame({"a": [1, 2], "b": ["x", "y"], "c": [4.1, 2.3]})
        >>> df = nw.from_native(df_native)

        Let's select numeric dtypes and multiply each value by 2:

        >>> df.select(ncs.numeric() * 2).to_native()
        shape: (2, 2)
        ┌─────┬─────┐
        │ a   ┆ c   │
        │ --- ┆ --- │
        │ i64 ┆ f64 │
        ╞═════╪═════╡
        │ 2   ┆ 8.2 │
        │ 4   ┆ 4.6 │
        └─────┴─────┘
    """
    return Selector(ExprNode(ExprKind.SELECTOR, "selectors.numeric"))


def boolean() -> Selector:
    """Select boolean columns.

    Examples:
        >>> import polars as pl
        >>> import narwhals as nw
        >>> import narwhals.selectors as ncs
        >>> df_native = pl.DataFrame({"a": [1, 2], "b": ["x", "y"], "c": [False, True]})
        >>> df = nw.from_native(df_native)

        Let's select boolean dtypes:

        >>> df.select(ncs.boolean())
        ┌──────────────────┐
        |Narwhals DataFrame|
        |------------------|
        |  shape: (2, 1)   |
        |  ┌───────┐       |
        |  │ c     │       |
        |  │ ---   │       |
        |  │ bool  │       |
        |  ╞═══════╡       |
        |  │ false │       |
        |  │ true  │       |
        |  └───────┘       |
        └──────────────────┘
    """
    return Selector(ExprNode(ExprKind.SELECTOR, "selectors.boolean"))


def string() -> Selector:
    """Select string columns.

    Examples:
        >>> import polars as pl
        >>> import narwhals as nw
        >>> import narwhals.selectors as ncs
        >>> df_native = pl.DataFrame({"a": [1, 2], "b": ["x", "y"], "c": [False, True]})
        >>> df = nw.from_native(df_native)

        Let's select string dtypes:

        >>> df.select(ncs.string()).to_native()
        shape: (2, 1)
        ┌─────┐
        │ b   │
        │ --- │
        │ str │
        ╞═════╡
        │ x   │
        │ y   │
        └─────┘
    """
    return Selector(ExprNode(ExprKind.SELECTOR, "selectors.string"))


def categorical() -> Selector:
    """Select categorical columns.

    Examples:
        >>> import polars as pl
        >>> import narwhals as nw
        >>> import narwhals.selectors as ncs
        >>> df_native = pl.DataFrame({"a": [1, 2], "b": ["x", "y"], "c": [False, True]})

        Let's convert column "b" to categorical, and then select categorical dtypes:

        >>> df = nw.from_native(df_native).with_columns(
        ...     b=nw.col("b").cast(nw.Categorical())
        ... )
        >>> df.select(ncs.categorical()).to_native()
        shape: (2, 1)
        ┌─────┐
        │ b   │
        │ --- │
        │ cat │
        ╞═════╡
        │ x   │
        │ y   │
        └─────┘
    """
    return Selector(ExprNode(ExprKind.SELECTOR, "selectors.categorical"))


def all() -> Selector:
    """Select all columns.

    Examples:
        >>> import pandas as pd
        >>> import narwhals as nw
        >>> import narwhals.selectors as ncs
        >>> df_native = pd.DataFrame({"a": [1, 2], "b": ["x", "y"], "c": [False, True]})
        >>> df = nw.from_native(df_native)

        Let's select all dtypes:

        >>> df.select(ncs.all()).to_native()
           a  b      c
        0  1  x  False
        1  2  y   True
    """
    return Selector(ExprNode(ExprKind.SELECTOR, "selectors.all"))


def datetime(
    time_unit: TimeUnit | Iterable[TimeUnit] | None = None,
    time_zone: str | timezone | Iterable[str | timezone | None] | None = ("*", None),
) -> Selector:
    """Select all datetime columns, optionally filtering by time unit/zone.

    Arguments:
        time_unit: One (or more) of the allowed timeunit precision strings, "ms", "us",
            "ns" and "s". Omit to select columns with any valid timeunit.
        time_zone: Specify which timezone(s) to select

            * One or more timezone strings, as defined in zoneinfo (to see valid options
                run `import zoneinfo; zoneinfo.available_timezones()` for a full list).
            * Set `None` to select Datetime columns that do not have a timezone.
            * Set `"*"` to select Datetime columns that have *any* timezone.

    Examples:
        >>> from datetime import datetime, timezone
        >>> import pyarrow as pa
        >>> import narwhals as nw
        >>> import narwhals.selectors as ncs
        >>>
        >>> utc_tz = timezone.utc
        >>> data = {
        ...     "tstamp_utc": [
        ...         datetime(2023, 4, 10, 12, 14, 16, 999000, tzinfo=utc_tz),
        ...         datetime(2025, 8, 25, 14, 18, 22, 666000, tzinfo=utc_tz),
        ...     ],
        ...     "tstamp": [
        ...         datetime(2000, 11, 20, 18, 12, 16, 600000),
        ...         datetime(2020, 10, 30, 10, 20, 25, 123000),
        ...     ],
        ...     "numeric": [3.14, 6.28],
        ... }
        >>> df_native = pa.table(data)
        >>> df_nw = nw.from_native(df_native)
        >>> df_nw.select(ncs.datetime()).to_native()
        pyarrow.Table
        tstamp_utc: timestamp[us, tz=UTC]
        tstamp: timestamp[us]
        ----
        tstamp_utc: [[2023-04-10 12:14:16.999000Z,2025-08-25 14:18:22.666000Z]]
        tstamp: [[2000-11-20 18:12:16.600000,2020-10-30 10:20:25.123000]]

        Select only datetime columns that have any time_zone specification:

        >>> df_nw.select(ncs.datetime(time_zone="*")).to_native()
        pyarrow.Table
        tstamp_utc: timestamp[us, tz=UTC]
        ----
        tstamp_utc: [[2023-04-10 12:14:16.999000Z,2025-08-25 14:18:22.666000Z]]
    """
    return Selector(
        ExprNode(
            ExprKind.SELECTOR,
            "selectors.datetime",
            time_unit=time_unit,
            time_zone=time_zone,
        )
    )


__all__ = [
    "all",
    "boolean",
    "by_dtype",
    "categorical",
    "datetime",
    "matches",
    "numeric",
    "string",
]



from __future__ import annotations

import narwhals as nw
from tests.utils import Constructor, assert_equal_data


def test_rename(constructor: Constructor) -> None:
    data = {"a": [1, 3, 2], "b": [4, 4, 6], "z": [7.0, 8.0, 9.0]}
    df = nw.from_native(constructor(data))
    result = df.rename({"a": "foo-bar", "b": "foo bar", "z": "z.b"})
    expected = {"foo-bar": [1, 3, 2], "foo bar": [4, 4, 6], "z.b": [7.0, 8.0, 9.0]}
    assert_equal_data(result, expected)
    result = df.rename({"b": "z", "z": "y"})
    expected = {"a": [1, 3, 2], "z": [4, 4, 6], "y": [7.0, 8.0, 9.0]}
    assert_equal_data(result, expected)



from __future__ import annotations

from contextlib import nullcontext as does_not_raise

import pytest

import narwhals as nw
from narwhals.exceptions import InvalidOperationError
from tests.utils import Constructor, ConstructorEager, assert_equal_data

data = {"a": [1, 1, None, 2]}
data_str = {"a": ["x", "x", "y", None]}


def test_unique_expr(constructor: Constructor) -> None:
    df = nw.from_native(constructor(data))
    context = (
        pytest.raises((InvalidOperationError, NotImplementedError))
        if isinstance(df, nw.LazyFrame)
        else does_not_raise()
    )
    with context:
        result = df.select(nw.col("a").unique()).sort("a")
        expected = {"a": [None, 1, 2]}
        assert_equal_data(result, expected)


def test_unique_expr_agg(
    constructor: Constructor, request: pytest.FixtureRequest
) -> None:
    if any(x in str(constructor) for x in ("duckdb", "pyspark", "ibis")):
        request.applymarker(pytest.mark.xfail)
    df = nw.from_native(constructor(data))
    result = df.select(nw.col("a").unique().sum())
    expected = {"a": [3]}
    assert_equal_data(result, expected)
    result = df.drop_nulls().select(nw.col("a").unique().len())
    expected = {"a": [2]}
    assert_equal_data(result, expected)


def test_unique_illegal_combination(constructor: Constructor) -> None:
    df = nw.from_native(constructor(data))
    with pytest.raises((InvalidOperationError, NotImplementedError)):
        df.select((nw.col("a").unique() + nw.col("a").unique()).sum())
    with pytest.raises((InvalidOperationError, NotImplementedError)):
        df.select(nw.col("a").unique() + nw.col("a"))


def test_unique_series(constructor_eager: ConstructorEager) -> None:
    series = nw.from_native(constructor_eager(data_str), eager_only=True)["a"]
    result = series.unique(maintain_order=True)
    expected = {"a": ["x", "y", None]}
    assert_equal_data({"a": result}, expected)
    result = series.drop_nulls().unique(maintain_order=True)
    expected = {"a": ["x", "y"]}
    assert_equal_data({"a": result}, expected)


def test_unique_series_numeric(constructor_eager: ConstructorEager) -> None:
    series = nw.from_native(constructor_eager(data), eager_only=True)["a"]
    result = series.unique(maintain_order=True)
    expected = {"a": [1, None, 2]}
    assert_equal_data({"a": result}, expected)
    result = series.drop_nulls().unique(maintain_order=True)
    expected = {"a": [1, 2]}
    assert_equal_data({"a": result}, expected)



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Callable, cast

import pandas as pd

from narwhals._compliant import DepthTrackingExpr, LazyExpr
from narwhals._dask.expr_dt import DaskExprDateTimeNamespace
from narwhals._dask.expr_str import DaskExprStringNamespace
from narwhals._dask.utils import (
    add_row_index,
    align_series_full_broadcast,
    make_group_by_kwargs,
    narwhals_to_native_dtype,
)
from narwhals._expression_parsing import evaluate_nodes, evaluate_output_names_and_aliases
from narwhals._pandas_like.expr import window_kwargs_to_pandas_equivalent
from narwhals._pandas_like.utils import get_dtype_backend, native_to_narwhals_dtype
from narwhals._utils import (
    Implementation,
    generate_temporary_column_name,
    no_default,
    not_implemented,
)
from narwhals.exceptions import InvalidOperationError

if TYPE_CHECKING:
    from collections.abc import Sequence

    import dask.dataframe.dask_expr as dx
    from typing_extensions import Self

    from narwhals._compliant.typing import (
        AliasNames,
        EvalNames,
        EvalSeries,
        NarwhalsAggregation,
    )
    from narwhals._dask.dataframe import DaskLazyFrame
    from narwhals._dask.namespace import DaskNamespace
    from narwhals._typing import NoDefault
    from narwhals._utils import Version, _LimitedContext
    from narwhals.typing import (
        FillNullStrategy,
        IntoDType,
        ModeKeepStrategy,
        RollingInterpolationMethod,
    )


def simple_aggregation(attr: str) -> Any:
    # Translate an aggregation which differs from Dask in name only.
    return lambda self, **kwargs: self._with_callable(
        lambda expr: getattr(expr, attr)(**kwargs).to_series()
    )


def simple_method(attr: str) -> Any:
    # Translate a method which differs from Dask in name only.
    return lambda self, **kwargs: self._with_callable(
        lambda expr: getattr(expr, attr)(**kwargs)
    )


def simple_binary(attr: str) -> Any:
    # Translate a binary method which differs from Dask in name only.
    return lambda self, other: self._binary_op(attr, other)


def trivial_binary_right(op: Callable[..., dx.Series]) -> Any:
    # Translate a (right) binary method which differs from Dask in name only.
    return lambda self, other: self._reverse_binary_op(op, other)


class DaskExpr(
    LazyExpr["DaskLazyFrame", "dx.Series"],  # pyright: ignore[reportInvalidTypeArguments]
    DepthTrackingExpr["DaskLazyFrame", "dx.Series"],  # pyright: ignore[reportInvalidTypeArguments]
):
    _implementation: Implementation = Implementation.DASK

    # Methods which are simple Narwhals->Dask translations.
    # Keep this simple, resist the temptation to do anything complex or clever.
    __add__ = simple_binary("__add__")
    __sub__ = simple_binary("__sub__")
    __mul__ = simple_binary("__mul__")
    __truediv__ = simple_binary("__truediv__")
    __pow__ = simple_binary("__pow__")
    __mod__ = simple_binary("__mod__")
    __eq__ = simple_binary("__eq__")
    __ne__ = simple_binary("__ne__")
    __gt__ = simple_binary("__gt__")
    __ge__ = simple_binary("__ge__")
    __lt__ = simple_binary("__lt__")
    __le__ = simple_binary("__le__")
    __and__ = simple_binary("__and__")
    __or__ = simple_binary("__or__")
    __rsub__ = trivial_binary_right(lambda x, y: x - y)
    __rtruediv__ = trivial_binary_right(lambda x, y: x / y)
    __rpow__ = trivial_binary_right(lambda x, y: x**y)
    __rmod__ = trivial_binary_right(lambda x, y: x % y)
    all = simple_aggregation("all")
    any = simple_aggregation("any")
    count = simple_aggregation("count")
    kurtosis = simple_aggregation("kurtosis")
    max = simple_aggregation("max")
    mean = simple_aggregation("mean")
    min = simple_aggregation("min")
    skew = simple_aggregation("skew")
    std = simple_aggregation("std")
    sum = simple_aggregation("sum")
    var = simple_aggregation("var")
    __invert__ = simple_method("__invert__")
    abs = simple_method("abs")
    diff = simple_method("diff")
    drop_nulls = simple_method("dropna")
    is_null = simple_method("isna")
    round = simple_method("round")
    unique = simple_method("unique")

    def __init__(
        self,
        call: EvalSeries[DaskLazyFrame, dx.Series],  # pyright: ignore[reportInvalidTypeForm]
        *,
        evaluate_output_names: EvalNames[DaskLazyFrame],
        alias_output_names: AliasNames | None,
        version: Version,
    ) -> None:
        self._call = call
        self._evaluate_output_names = evaluate_output_names
        self._alias_output_names = alias_output_names
        self._version = version

    def __call__(self, df: DaskLazyFrame) -> Sequence[dx.Series]:
        return self._call(df)

    def __narwhals_namespace__(self) -> DaskNamespace:  # pragma: no cover
        from narwhals._dask.namespace import DaskNamespace

        return DaskNamespace(version=self._version)

    def broadcast(self) -> Self:
        def func(df: DaskLazyFrame) -> list[dx.Series]:
            # result.loc[0][0] is a workaround for dask~<=2024.10.0/dask_expr~<=1.1.16
            #   that raised a KeyError for result[0] during collection.
            return [result.loc[0][0] for result in self(df)]

        return self.__class__(
            func,
            evaluate_output_names=self._evaluate_output_names,
            alias_output_names=self._alias_output_names,
            version=self._version,
        )

    @classmethod
    def from_column_names(
        cls: type[Self],
        evaluate_column_names: EvalNames[DaskLazyFrame],
        /,
        *,
        context: _LimitedContext,
    ) -> Self:
        def func(df: DaskLazyFrame) -> list[dx.Series]:
            try:
                return [
                    df._native_frame[column_name]
                    for column_name in evaluate_column_names(df)
                ]
            except KeyError as e:
                if error := df._check_columns_exist(evaluate_column_names(df)):
                    raise error from e
                raise

        return cls(
            func,
            evaluate_output_names=evaluate_column_names,
            alias_output_names=None,
            version=context._version,
        )

    @classmethod
    def from_column_indices(cls, *column_indices: int, context: _LimitedContext) -> Self:
        def func(df: DaskLazyFrame) -> list[dx.Series]:
            return [df.native.iloc[:, i] for i in column_indices]

        return cls(
            func,
            evaluate_output_names=cls._eval_names_indices(column_indices),
            alias_output_names=None,
            version=context._version,
        )

    def _with_callable(
        self,
        # First argument to `call` should be `dx.Series`
        call: Callable[..., dx.Series],
        /,
        **expressifiable_args: Self,
    ) -> Self:
        def func(df: DaskLazyFrame) -> list[dx.Series]:
            native_results: list[dx.Series] = []
            native_series_list = self._call(df)
            other_native_series = {
                key: df._evaluate_single_output_expr(value)
                for key, value in expressifiable_args.items()
            }
            for native_series in native_series_list:
                result_native = call(native_series, **other_native_series)
                native_results.append(result_native)
            return native_results

        return self.__class__(
            func,
            evaluate_output_names=self._evaluate_output_names,
            alias_output_names=self._alias_output_names,
            version=self._version,
        )

    def _with_alias_output_names(self, func: AliasNames | None, /) -> Self:
        current_alias_output_names = self._alias_output_names
        alias_output_names = (
            None
            if func is None
            else func
            if current_alias_output_names is None
            else lambda output_names: func(current_alias_output_names(output_names))
        )
        return type(self)(
            call=self._call,
            evaluate_output_names=self._evaluate_output_names,
            alias_output_names=alias_output_names,
            version=self._version,
        )

    def _with_binary(
        self, call: Callable[[dx.Series, Any], dx.Series], other: Any
    ) -> Self:
        return self._with_callable(call, other=other)

    def _binary_op(self, op_name: str, other: Any) -> Self:
        return self._with_binary(lambda expr, other: getattr(expr, op_name)(other), other)

    def _reverse_binary_op(
        self, operator_func: Callable[..., dx.Series], other: Any
    ) -> Self:
        return self._with_binary(
            lambda expr, other: operator_func(other, expr), other
        ).alias("literal")

    def __floordiv__(self, other: Any) -> Self:
        def _floordiv(
            df: DaskLazyFrame, series: dx.Series, other: dx.Series
        ) -> dx.Series:
            series, other = align_series_full_broadcast(df, series, other)
            return (series.__floordiv__(other)).where(other != 0, None)

        def func(df: DaskLazyFrame) -> list[dx.Series]:
            other_series = df._evaluate_single_output_expr(other)
            return [_floordiv(df, series, other_series) for series in self(df)]

        return self.__class__(
            func,
            evaluate_output_names=self._evaluate_output_names,
            alias_output_names=self._alias_output_names,
            version=self._version,
        )

    def __rfloordiv__(self, other: Any) -> Self:
        def _rfloordiv(
            df: DaskLazyFrame, series: dx.Series, other: dx.Series
        ) -> dx.Series:
            series, other = align_series_full_broadcast(df, series, other)
            return (other.__floordiv__(series)).where(series != 0, None)

        def func(df: DaskLazyFrame) -> list[dx.Series]:
            other_native = df._evaluate_single_output_expr(other)
            return [_rfloordiv(df, series, other_native) for series in self(df)]

        return self.__class__(
            func,
            evaluate_output_names=self._evaluate_output_names,
            alias_output_names=self._alias_output_names,
            version=self._version,
        ).alias("literal")

    def median(self) -> Self:
        from narwhals.exceptions import InvalidOperationError

        def func(s: dx.Series) -> dx.Series:
            dtype = native_to_narwhals_dtype(s.dtype, self._version, Implementation.DASK)
            if not dtype.is_numeric():
                msg = "`median` operation not supported for non-numeric input type."
                raise InvalidOperationError(msg)
            return s.median_approximate().to_series()

        return self._with_callable(func)

    def shift(self, n: int) -> Self:
        return self._with_callable(lambda expr: expr.shift(n))

    def cum_sum(self, *, reverse: bool) -> Self:
        if reverse:  # pragma: no cover
            # https://github.com/dask/dask/issues/11802
            msg = "`cum_sum(reverse=True)` is not supported with Dask backend"
            raise NotImplementedError(msg)

        return self._with_callable(lambda expr: expr.cumsum())

    def cum_count(self, *, reverse: bool) -> Self:
        if reverse:  # pragma: no cover
            msg = "`cum_count(reverse=True)` is not supported with Dask backend"
            raise NotImplementedError(msg)

        return self._with_callable(lambda expr: (~expr.isna()).astype(int).cumsum())

    def cum_min(self, *, reverse: bool) -> Self:
        if reverse:  # pragma: no cover
            msg = "`cum_min(reverse=True)` is not supported with Dask backend"
            raise NotImplementedError(msg)

        return self._with_callable(lambda expr: expr.cummin())

    def cum_max(self, *, reverse: bool) -> Self:
        if reverse:  # pragma: no cover
            msg = "`cum_max(reverse=True)` is not supported with Dask backend"
            raise NotImplementedError(msg)

        return self._with_callable(lambda expr: expr.cummax())

    def cum_prod(self, *, reverse: bool) -> Self:
        if reverse:  # pragma: no cover
            msg = "`cum_prod(reverse=True)` is not supported with Dask backend"
            raise NotImplementedError(msg)

        return self._with_callable(lambda expr: expr.cumprod())

    def rolling_sum(self, window_size: int, *, min_samples: int, center: bool) -> Self:
        return self._with_callable(
            lambda expr: expr.rolling(
                window=window_size, min_periods=min_samples, center=center
            ).sum()
        )

    def rolling_mean(self, window_size: int, *, min_samples: int, center: bool) -> Self:
        return self._with_callable(
            lambda expr: expr.rolling(
                window=window_size, min_periods=min_samples, center=center
            ).mean()
        )

    def rolling_var(
        self, window_size: int, *, min_samples: int, center: bool, ddof: int
    ) -> Self:
        if ddof == 1:
            return self._with_callable(
                lambda expr: expr.rolling(
                    window=window_size, min_periods=min_samples, center=center
                ).var()
            )
        msg = "Dask backend only supports `ddof=1` for `rolling_var`"
        raise NotImplementedError(msg)

    def rolling_std(
        self, window_size: int, *, min_samples: int, center: bool, ddof: int
    ) -> Self:
        if ddof == 1:
            return self._with_callable(
                lambda expr: expr.rolling(
                    window=window_size, min_periods=min_samples, center=center
                ).std()
            )
        msg = "Dask backend only supports `ddof=1` for `rolling_std`"
        raise NotImplementedError(msg)

    def floor(self) -> Self:
        import dask.array as da

        return self._with_callable(da.floor)

    def ceil(self) -> Self:
        import dask.array as da

        return self._with_callable(da.ceil)

    def fill_nan(self, value: float | None) -> Self:
        value_nullable = pd.NA if value is None else value
        value_numpy = float("nan") if value is None else value

        def func(expr: dx.Series) -> dx.Series:
            # If/when pandas exposes an API which distinguishes NaN vs null, use that.
            mask = cast("dx.Series", expr != expr)  # noqa: PLR0124
            mask = mask.fillna(False)
            fill = (
                value_nullable
                if get_dtype_backend(expr.dtype, self._implementation)
                else value_numpy
            )
            return expr.mask(mask, fill)  # pyright: ignore[reportArgumentType]

        return self._with_callable(func)

    def fill_null(
        self, value: Self | None, strategy: FillNullStrategy | None, limit: int | None
    ) -> Self:
        def func(expr: dx.Series) -> dx.Series:
            if value is not None:
                res_ser = expr.fillna(value)
            else:
                res_ser = (
                    expr.ffill(limit=limit)
                    if strategy == "forward"
                    else expr.bfill(limit=limit)
                )
            return res_ser

        return self._with_callable(func)

    def clip(self, lower_bound: Self, upper_bound: Self) -> Self:
        return self._with_callable(
            lambda expr, lower_bound, upper_bound: expr.clip(
                lower=lower_bound, upper=upper_bound
            ),
            lower_bound=lower_bound,
            upper_bound=upper_bound,
        )

    def clip_lower(self, lower_bound: Self) -> Self:
        return self._with_callable(
            lambda expr, lower_bound: expr.clip(lower=lower_bound),
            lower_bound=lower_bound,
        )

    def clip_upper(self, upper_bound: Self) -> Self:
        return self._with_callable(
            lambda expr, upper_bound: expr.clip(upper=upper_bound),
            upper_bound=upper_bound,
        )

    def n_unique(self) -> Self:
        return self._with_callable(lambda expr: expr.nunique(dropna=False).to_series())

    def is_nan(self) -> Self:
        def func(expr: dx.Series) -> dx.Series:
            dtype = native_to_narwhals_dtype(
                expr.dtype, self._version, self._implementation
            )
            if dtype.is_numeric():
                return expr != expr  # pyright: ignore[reportReturnType] # noqa: PLR0124
            msg = f"`.is_nan` only supported for numeric dtypes and not {dtype}, did you mean `.is_null`?"
            raise InvalidOperationError(msg)

        return self._with_callable(func)

    def len(self) -> Self:
        return self._with_callable(lambda expr: expr.size.to_series())

    def quantile(
        self, quantile: float, interpolation: RollingInterpolationMethod
    ) -> Self:
        if interpolation == "linear":

            def func(expr: dx.Series) -> dx.Series:
                if expr.npartitions > 1:
                    msg = "`Expr.quantile` is not supported for Dask backend with multiple partitions."
                    raise NotImplementedError(msg)
                return expr.quantile(
                    q=quantile, method="dask"
                ).to_series()  # pragma: no cover

            return self._with_callable(func)
        msg = "`higher`, `lower`, `midpoint`, `nearest` - interpolation methods are not supported by Dask. Please use `linear` instead."
        raise NotImplementedError(msg)

    def is_first_distinct(self) -> Self:
        def func(expr: dx.Series) -> dx.Series:
            _name = expr.name
            col_token = generate_temporary_column_name(
                n_bytes=8, columns=[_name], prefix="row_index_"
            )
            frame = add_row_index(expr.to_frame(), col_token)
            group_by_kwargs = make_group_by_kwargs(drop_null_keys=False)
            first_distinct_index = frame.groupby(_name, **group_by_kwargs).agg(
                {col_token: "min"}
            )[col_token]
            return frame[col_token].isin(first_distinct_index)

        return self._with_callable(func)

    def is_last_distinct(self) -> Self:
        def func(expr: dx.Series) -> dx.Series:
            _name = expr.name
            col_token = generate_temporary_column_name(
                n_bytes=8, columns=[_name], prefix="row_index_"
            )
            frame = add_row_index(expr.to_frame(), col_token)
            group_by_kwargs = make_group_by_kwargs(drop_null_keys=False)
            last_distinct_index = frame.groupby(_name, **group_by_kwargs).agg(
                {col_token: "max"}
            )[col_token]
            return frame[col_token].isin(last_distinct_index)

        return self._with_callable(func)

    def is_unique(self) -> Self:
        def func(expr: dx.Series) -> dx.Series:
            _name = expr.name
            group_by_kwargs = make_group_by_kwargs(drop_null_keys=False)
            return (
                expr.to_frame()
                .groupby(_name, **group_by_kwargs)
                .transform("size", meta=(_name, int))
                == 1
            )

        return self._with_callable(func)

    def is_in(self, other: Any) -> Self:
        return self._with_callable(lambda expr: expr.isin(other))

    def null_count(self) -> Self:
        return self._with_callable(lambda expr: expr.isna().sum().to_series())

    def _over_without_partition_by(self, order_by: Sequence[str]) -> Self:
        # This is something like `nw.col('a').cum_sum().order_by(key)`
        # which we can always easily support, as it doesn't require grouping.
        def func(df: DaskLazyFrame) -> Sequence[dx.Series]:
            return self(df.sort(*order_by, descending=False, nulls_last=False))

        return self.__class__(
            func,
            evaluate_output_names=self._evaluate_output_names,
            alias_output_names=self._alias_output_names,
            version=self._version,
        )

    def over(self, partition_by: Sequence[str], order_by: Sequence[str]) -> Self:
        if not partition_by:
            assert order_by  # noqa: S101
            return self._over_without_partition_by(order_by)
        # pandas is a required dependency of dask so it's safe to import this
        from narwhals._pandas_like.group_by import PandasLikeGroupBy

        # We have something like prev.leaf().over(...) (e.g. `nw.col('a').sum().over('b')`), where:
        # - `prev` must be elementwise (in the example: `nw.col('a')`)
        # - `leaf` must be a "simple" function, i.e. one that pandas supports in `transform`
        #   (in the example: `sum`)
        #
        # We first evaluate `prev` as-is, and then evaluate `leaf().over(...)`` by using `transform`
        # or other DataFrameGroupBy methods.
        meta = self._metadata
        if partition_by and (
            meta.prev is not None and not meta.prev.is_elementwise
        ):  # pragma: no cover
            msg = (
                "Only elementary expressions are supported for `.over` in dask backend "
                "when `partition_by` is specified.\n\n"
                "Please see: "
                "https://narwhals-dev.github.io/narwhals/concepts/improve_group_by_operation/"
            )
            raise NotImplementedError(msg)

        if order_by:
            # Wrong results https://github.com/dask/dask/issues/11806.
            msg = "`over` with `order_by` is not yet supported in Dask."
            raise NotImplementedError(msg)

        nodes = list(reversed(list(self._metadata.iter_nodes_reversed())))
        leaf_node = nodes[-1]
        function_name = cast("NarwhalsAggregation", leaf_node.name)
        try:
            dask_function_name = PandasLikeGroupBy._REMAP_AGGS[function_name]
        except KeyError:
            # window functions are unsupported: https://github.com/dask/dask/issues/11806
            msg = (
                f"Unsupported function: {function_name} in `over` context.\n\n"
                f"Supported functions are {', '.join(PandasLikeGroupBy._REMAP_AGGS)}\n"
            )
            raise NotImplementedError(msg) from None
        dask_kwargs = window_kwargs_to_pandas_equivalent(function_name, leaf_node.kwargs)

        def func(df: DaskLazyFrame) -> Sequence[dx.Series]:
            plx = self.__narwhals_namespace__()
            if meta.prev is not None:
                df = df.with_columns(cast("DaskExpr", evaluate_nodes(nodes[:-1], plx)))
            _, aliases = evaluate_output_names_and_aliases(self, df, [])

            with warnings.catch_warnings():
                # https://github.com/dask/dask/issues/11804
                warnings.filterwarnings(
                    "ignore", message=".*`meta` is not specified", category=UserWarning
                )
                group_by_kwargs = make_group_by_kwargs(drop_null_keys=False)
                grouped = df.native.groupby(partition_by, **group_by_kwargs)
                if dask_function_name == "size":
                    if len(aliases) != 1:  # pragma: no cover
                        msg = "Safety check failed, please report a bug."
                        raise AssertionError(msg)
                    res_native = grouped.transform(
                        dask_function_name, **dask_kwargs
                    ).to_frame(aliases[0])
                else:
                    res_native = grouped[list(aliases)].transform(
                        dask_function_name, **dask_kwargs
                    )
            result_frame = df._with_native(res_native).native
            return [result_frame[name] for name in aliases]

        return self.__class__(
            func,
            evaluate_output_names=self._evaluate_output_names,
            alias_output_names=self._alias_output_names,
            version=self._version,
        )

    def cast(self, dtype: IntoDType) -> Self:
        def func(expr: dx.Series) -> dx.Series:
            native_dtype = narwhals_to_native_dtype(dtype, self._version)
            return expr.astype(native_dtype)

        return self._with_callable(func)

    def is_finite(self) -> Self:
        import dask.array as da

        return self._with_callable(da.isfinite)

    def log(self, base: float) -> Self:
        import dask.array as da

        def _log(expr: dx.Series) -> dx.Series:
            return da.log(expr) / da.log(base)

        return self._with_callable(_log)

    def exp(self) -> Self:
        import dask.array as da

        return self._with_callable(da.exp)

    def sin(self) -> Self:
        import dask.array as da

        return self._with_callable(da.sin)

    def cos(self) -> Self:
        import dask.array as da

        return self._with_callable(da.cos)

    def sqrt(self) -> Self:
        import dask.array as da

        return self._with_callable(da.sqrt)

    def mode(self, *, keep: ModeKeepStrategy) -> Self:
        def func(expr: dx.Series) -> dx.Series:
            _name = expr.name
            result = expr.to_frame().mode()[_name]
            return result.head(1) if keep == "any" else result

        return self._with_callable(func)

    def replace_strict(
        self,
        default: DaskExpr | NoDefault,
        old: Sequence[Any],
        new: Sequence[Any],
        *,
        return_dtype: IntoDType | None,
    ) -> Self:
        if default is no_default:
            msg = "`replace_strict` requires an explicit value for `default` for dask backend."
            raise ValueError(msg)

        mapping = dict(zip(old, new))
        old_ = list(old)

        def func(df: DaskLazyFrame) -> list[dx.Series]:
            default_series = df._evaluate_single_output_expr(default)
            results = [
                series.replace(mapping).where(series.isin(old_), default_series)
                for series in self(df)
            ]

            if return_dtype:
                native_dtype = narwhals_to_native_dtype(return_dtype, self._version)
                return [res.astype(native_dtype) for res in results]

            return results

        return self.__class__(
            func,
            evaluate_output_names=self._evaluate_output_names,
            alias_output_names=self._alias_output_names,
            version=self._version,
        )

    @property
    def str(self) -> DaskExprStringNamespace:
        return DaskExprStringNamespace(self)

    @property
    def dt(self) -> DaskExprDateTimeNamespace:
        return DaskExprDateTimeNamespace(self)

    any_value = not_implemented()
    filter = not_implemented()
    first = not_implemented()
    rank = not_implemented()
    last = not_implemented()

    # namespaces
    list: not_implemented = not_implemented()  # type: ignore[assignment]
    struct: not_implemented = not_implemented()  # type: ignore[assignment]



from __future__ import annotations

import pytest

import narwhals as nw
from narwhals.exceptions import InvalidOperationError
from tests.utils import DUCKDB_VERSION, POLARS_VERSION, Constructor, assert_equal_data


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        (nw.col("a"), [-1, 2, 3]),
        (nw.col("a").mean(), [4 / 3, 4 / 3, 4 / 3]),
        (nw.col("a").cum_sum().over(order_by="i"), [-1, 1, 4]),
        (nw.col("a").alias("b").cum_sum().over(order_by="i"), [-1, 1, 4]),
        (nw.col("a").cum_sum().abs().over(order_by="i"), [1, 1, 4]),
        ((nw.col("a").cum_sum() + 1).over(order_by="i"), [0, 2, 5]),
        (
            nw.sum_horizontal(nw.col("a"), nw.col("a").cum_sum()).over(order_by="a"),
            [-2, 3, 7],
        ),
        (
            nw.sum_horizontal(nw.col("a"), nw.col("a").cum_sum().over(order_by="i")),
            [-2, 3, 7],
        ),
        (
            nw.sum_horizontal(nw.col("a").diff(), nw.col("a").cum_sum()).over(
                order_by="i"
            ),
            [-1.0, 4.0, 5.0],
        ),
        (
            nw.sum_horizontal(nw.col("a").diff().abs(), nw.col("a").cum_sum()).over(
                order_by="i"
            ),
            [-1.0, 4.0, 5.0],
        ),
        (
            (nw.col("a").sum() + nw.col("a").rolling_sum(2, min_samples=1)).over(
                order_by="i"
            ),
            [3.0, 5.0, 9.0],
        ),
        ((nw.col("a").sum() + nw.col("a").mean()).over("b"), [1.5, 1.5, 6.0]),
        (
            (nw.col("a").mean().abs() + nw.sum_horizontal(nw.col("a").diff())).over(
                order_by="i"
            ),
            [4 / 3, 13 / 3, 7 / 3],
        ),
    ],
)
def test_over_pushdown(
    constructor: Constructor, expr: nw.Expr, expected: list[float]
) -> None:
    if "polars" in str(constructor) and POLARS_VERSION < (1, 10):
        pytest.skip()
    if "duckdb" in str(constructor) and DUCKDB_VERSION < (1, 3):
        pytest.skip()
    data = {"a": [-1, 2, 3], "b": [1, 1, 2], "i": [0, 1, 2]}
    df = nw.from_native(constructor(data)).lazy()
    result = df.select("i", a=expr).sort("i").select("a")
    assert_equal_data(result, {"a": expected})


@pytest.mark.parametrize(
    ("expr", "expected"), [((nw.col("a") - nw.col("a").mean()).over("b"), [-1.5, 1.5, 0])]
)
def test_per_group_broadcasting(
    constructor: Constructor,
    expr: nw.Expr,
    expected: list[float],
    request: pytest.FixtureRequest,
) -> None:
    if "dask" in str(constructor):
        # sigh...
        request.applymarker(pytest.mark.xfail)
    if "duckdb" in str(constructor) and DUCKDB_VERSION < (1, 3):
        pytest.skip()
    data = {"a": [-1, 2, 3], "b": [1, 1, 2], "i": [0, 1, 2]}
    df = nw.from_native(constructor(data)).lazy()
    result = df.select("i", a=expr).sort("i").select("a")
    assert_equal_data(result, {"a": expected})


@pytest.mark.parametrize(
    "expr",
    [
        nw.col("a").cum_sum(),
        nw.col("a").cum_sum().cum_sum().over(order_by="i"),
        nw.col("a").cum_sum().cum_sum(),
        nw.sum_horizontal(nw.col("a"), nw.col("a").cum_sum()),
        nw.sum_horizontal(nw.col("a").diff(), nw.col("a").cum_sum().over(order_by="i")),
        nw.col("a").mean().over(order_by="i"),
        nw.col("a").mean().over("b").over("c"),
        nw.col("a").mean().over("b").over("c", order_by="i"),
        nw.col("a").mean().mean(),
        nw.col("a").mean().sum(),
        nw.col("a").mean().drop_nulls(),
        nw.col("a").mean().rank(),
        nw.col("a").mean().is_unique(),
        nw.col("a").mean().diff(),
        nw.col("a").drop_nulls().over("b"),
        nw.col("a").drop_nulls().over("b", order_by="i"),
        nw.col("a").diff().drop_nulls().over("b", order_by="i"),
        nw.col("a").filter(nw.col("b").sum().over("c") > 1).sum().over("d"),
    ],
)
def test_invalid_operations(constructor: Constructor, expr: nw.Expr) -> None:
    if "polars" in str(constructor) and POLARS_VERSION < (1, 10):
        pytest.skip()
    df = nw.from_native(
        constructor({"a": [-1, 2, 3], "b": [1, 1, 1], "c": [2, 2, 2], "i": [0, 1, 2]})
    ).lazy()
    with pytest.raises((InvalidOperationError, NotImplementedError)):
        df.select(a=expr)


def test_invalid_elementwise_over() -> None:
    # This one raises before it's even evaluated.
    with pytest.raises(InvalidOperationError):
        nw.col("a").fill_null(3).over("b")


def test_rank_with_order_by_pushdown() -> None:
    pytest.importorskip("pandas")
    import pandas as pd

    df = nw.from_native(pd.DataFrame({"a": [1, 1, 2], "i": [2, 1, 0]}))
    result = df.select(
        "a",
        res=nw.sum_horizontal(nw.col("a").rank("ordinal"), nw.lit(1)).over(order_by="i"),
    )
    expected = {"a": [1, 1, 2], "res": [3.0, 2.0, 4.0]}
    assert_equal_data(result, expected)



from __future__ import annotations

from inspect import getdoc
from typing import TYPE_CHECKING, Any

import pytest

import narwhals as nw
import narwhals.stable.v2 as nw_v2

if TYPE_CHECKING:
    from collections.abc import Iterator


def remove_docstring_examples(doc: str) -> str:
    if "Examples:" in doc:
        return doc[: doc.find("Examples:")].rstrip()
    return doc.rstrip()


def test_stable_api_completeness() -> None:
    v2_api = nw_v2.__all__
    main_namespace_api = nw.__all__
    extra = set(v2_api).difference(main_namespace_api)
    assert not extra
    missing = set(main_namespace_api).difference(v2_api).difference({"stable"})
    assert not missing


def test_stable_api_docstrings() -> None:
    main_namespace_api = nw.__all__
    for item in main_namespace_api:
        if item == "from_dict":
            # We keep `native_namespace` around in the main namespace
            # until at least hierarchical forecast make a release
            continue
        if (doc := getdoc(getattr(nw, item))) is None:
            continue
        v2_doc = getdoc(getattr(nw_v2, item))
        assert v2_doc is not None
        assert remove_docstring_examples(v2_doc) == remove_docstring_examples(doc), item


def _iter_api_method_docs(obj: Any, *exclude: str) -> Iterator[tuple[str, str]]:
    for name in dir(obj):
        if (
            not name.startswith("_")
            and name not in exclude
            and (doc := getdoc(getattr(obj, name)))
        ):
            yield name, doc


def test_dataframe_docstrings() -> None:
    pytest.importorskip("polars")
    import polars as pl

    df_v2 = nw_v2.from_native(pl.DataFrame())
    df = nw.from_native(pl.DataFrame())
    for method_name, doc in _iter_api_method_docs(df):
        doc_v2 = getdoc(getattr(df_v2, method_name))
        assert doc_v2
        assert remove_docstring_examples(doc_v2) == remove_docstring_examples(doc)


def test_lazyframe_docstrings() -> None:
    pytest.importorskip("polars")
    import polars as pl

    ldf_v2 = nw_v2.from_native(pl.LazyFrame())
    ldf = nw.from_native(pl.LazyFrame())
    performance_warning = {"schema", "columns"}
    deprecated = {"tail", "gather_every"}
    for method_name, doc in _iter_api_method_docs(ldf, *performance_warning, *deprecated):
        doc_v2 = getdoc(getattr(ldf_v2, method_name))
        assert doc_v2
        assert remove_docstring_examples(doc_v2) == remove_docstring_examples(doc)


def test_series_docstrings() -> None:
    pytest.importorskip("polars")
    import polars as pl

    ser_v2 = nw_v2.from_native(pl.Series(), series_only=True)
    ser = nw.from_native(pl.Series(), series_only=True)
    for method_name, doc in _iter_api_method_docs(ser):
        if method_name in "hist":
            # This is still very unstable in Polars so we don't have it in stable.v2 yet.
            continue
        doc_v2 = getdoc(getattr(ser_v2, method_name))
        assert doc_v2
        assert remove_docstring_examples(doc_v2) == remove_docstring_examples(doc)



from __future__ import annotations

import enum
from contextlib import AbstractContextManager, nullcontext as does_not_warn
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Any, Literal

import pytest

import narwhals as nw
from narwhals.exceptions import InvalidOperationError, PerformanceWarning
from tests.utils import (
    PANDAS_VERSION,
    POLARS_VERSION,
    PYARROW_VERSION,
    assert_equal_hash,
    pyspark_session,
)

if TYPE_CHECKING:
    from collections.abc import Iterable

    from narwhals.typing import IntoFrame, IntoSeries, NonNestedDType
    from tests.utils import Constructor, ConstructorPandasLike, NestedOrEnumDType


@pytest.mark.parametrize("time_unit", ["us", "ns", "ms"])
@pytest.mark.parametrize("time_zone", ["Europe/Rome", timezone.utc, None])
def test_datetime_valid(
    time_unit: Literal["us", "ns", "ms"], time_zone: str | timezone | None
) -> None:
    dtype = nw.Datetime(time_unit=time_unit, time_zone=time_zone)

    assert dtype == nw.Datetime(time_unit=time_unit, time_zone=time_zone)
    assert dtype == nw.Datetime

    if time_zone:
        assert dtype != nw.Datetime(time_unit=time_unit)
    if time_unit != "ms":
        assert dtype != nw.Datetime(time_unit="ms")


@pytest.mark.parametrize("time_unit", ["abc"])
def test_datetime_invalid(time_unit: str) -> None:
    with pytest.raises(ValueError, match="invalid `time_unit`"):
        nw.Datetime(time_unit=time_unit)  # type: ignore[arg-type]


@pytest.mark.parametrize("time_unit", ["us", "ns", "ms"])
def test_duration_valid(time_unit: Literal["us", "ns", "ms"]) -> None:
    dtype = nw.Duration(time_unit=time_unit)

    assert dtype == nw.Duration(time_unit=time_unit)
    assert dtype == nw.Duration

    if time_unit != "ms":
        assert dtype != nw.Duration(time_unit="ms")


@pytest.mark.parametrize("time_unit", ["abc"])
def test_duration_invalid(time_unit: str) -> None:
    with pytest.raises(ValueError, match="invalid `time_unit`"):
        nw.Duration(time_unit=time_unit)  # type: ignore[arg-type]


def test_list_valid() -> None:
    dtype = nw.List(nw.Int64)
    assert dtype == nw.List(nw.Int64)
    assert dtype == nw.List
    assert dtype != nw.List(nw.Float32)
    assert dtype != nw.Duration
    assert repr(dtype) == "List(Int64)"
    dtype = nw.List(nw.List(nw.Int64))
    assert dtype == nw.List(nw.List(nw.Int64))
    assert dtype == nw.List
    assert dtype != nw.List(nw.List(nw.Float32))
    assert_equal_hash(dtype, nw.List(nw.List(nw.Int64)))


def test_array_valid() -> None:
    dtype = nw.Array(nw.Int64, 2)
    assert dtype == nw.Array(nw.Int64, 2)
    assert dtype == nw.Array
    assert dtype != nw.Array(nw.Int64, 3)
    assert dtype != nw.Array(nw.Float32, 2)
    assert dtype != nw.Duration
    assert repr(dtype) == "Array(Int64, shape=(2,))"
    dtype = nw.Array(nw.Array(nw.Int64, 2), 2)
    assert dtype == nw.Array(nw.Array(nw.Int64, 2), 2)
    assert dtype == nw.Array
    assert dtype != nw.Array(nw.Array(nw.Float32, 2), 2)
    assert_equal_hash(dtype, nw.Array(nw.Array(nw.Int64, 2), 2))

    with pytest.raises(TypeError, match="invalid input for shape"):
        nw.Array(nw.Int64(), shape=None)  # type: ignore[arg-type]

    with pytest.raises(TypeError, match="invalid input for shape"):
        nw.Array(nw.Int64(), shape="invalid_type")  # type: ignore[arg-type]


def test_struct_valid() -> None:
    dtype = nw.Struct([nw.Field("a", nw.Int64)])
    assert dtype == nw.Struct([nw.Field("a", nw.Int64)])
    assert dtype == nw.Struct
    assert dtype != nw.Struct([nw.Field("a", nw.Float32)])
    assert dtype != nw.Duration
    assert repr(dtype) == "Struct({'a': Int64})"

    dtype = nw.Struct({"a": nw.Int64, "b": nw.String})
    assert dtype == nw.Struct({"a": nw.Int64, "b": nw.String})
    assert dtype.to_schema() == nw.Struct({"a": nw.Int64, "b": nw.String}).to_schema()
    assert dtype == nw.Struct
    assert dtype != nw.Struct({"a": nw.Int32, "b": nw.String})
    assert_equal_hash(dtype, nw.Struct({"a": nw.Int64, "b": nw.String}))


def test_struct_reverse() -> None:
    dtype1 = nw.Struct({"a": nw.Int64, "b": nw.String})
    dtype1_reversed = nw.Struct([nw.Field(*field) for field in reversed(dtype1)])
    dtype2 = nw.Struct({"b": nw.String, "a": nw.Int64})
    assert dtype1_reversed == dtype2


def test_field_repr() -> None:
    dtype = nw.Field("a", nw.Int32)
    assert repr(dtype) == "Field('a', Int32)"


def test_field_eq() -> None:
    field_1 = nw.Field("a", nw.String)
    field_2 = nw.Field("a", nw.String())
    field_3 = nw.Field("b", nw.Datetime())
    field_4 = nw.Field("b", nw.Datetime("ms"))
    field_5 = nw.Field("b", nw.Datetime)

    assert field_1 == field_2
    assert field_2 != field_3
    # bit of a head-scratcher
    assert field_3 != field_4
    assert field_3 == field_5
    assert field_4 == field_5
    assert field_1 != field_1.dtype


def test_struct_hashes() -> None:
    dtypes = (
        nw.Struct,
        nw.Struct([nw.Field("a", nw.Int64)]),
        nw.Struct([nw.Field("a", nw.Int64), nw.Field("b", nw.List(nw.Int64))]),
    )
    assert len({hash(tp) for tp in (dtypes)}) == 3


def test_2d_array(constructor: Constructor, request: pytest.FixtureRequest) -> None:
    version_conditions = [
        (PANDAS_VERSION < (2, 2), "Requires pandas 2.2+ for 2D array support"),
        (
            "pyarrow_table" in str(constructor) and PYARROW_VERSION < (14,),
            "PyArrow 14+ required for 2D array support",
        ),
    ]
    for condition, reason in version_conditions:
        if condition:
            pytest.skip(reason)
    if "pandas" in str(constructor):
        pytest.importorskip("pyarrow")

    if any(x in str(constructor) for x in ("dask", "cudf", "pyspark")):
        request.applymarker(
            pytest.mark.xfail(
                reason="2D array operations not supported in these backends"
            )
        )

    data = {"a": [[[1, 2], [3, 4], [5, 6]]]}
    df = nw.from_native(constructor(data)).with_columns(
        a=nw.col("a").cast(nw.Array(nw.Int64(), (3, 2)))
    )
    assert df.collect_schema()["a"] == nw.Array(nw.Int64(), (3, 2))
    assert df.collect_schema()["a"] == nw.Array(nw.Array(nw.Int64(), 2), 3)


def test_second_time_unit() -> None:
    pytest.importorskip("pandas")
    pytest.importorskip("pyarrow")
    import numpy as np
    import pandas as pd
    import pyarrow as pa

    s: IntoSeries = pd.Series(np.array([np.datetime64("2020-01-01", "s")]))
    result = nw.from_native(s, series_only=True)
    expected_unit: Literal["ns", "us", "ms", "s"] = (
        "s" if PANDAS_VERSION >= (2,) else "ns"
    )
    assert result.dtype == nw.Datetime(expected_unit)

    ts_sec = pa.timestamp("s")
    s = pa.chunked_array([pa.array([datetime(2020, 1, 1)], type=ts_sec)], type=ts_sec)
    result = nw.from_native(s, series_only=True)
    assert result.dtype == nw.Datetime("s")

    s = pd.Series(np.array([np.timedelta64(1, "s")]))
    result = nw.from_native(s, series_only=True)
    assert result.dtype == nw.Duration(expected_unit)

    dur_sec = pa.duration("s")
    s = pa.chunked_array([pa.array([timedelta(1)], type=dur_sec)], type=dur_sec)
    result = nw.from_native(s, series_only=True)
    assert result.dtype == nw.Duration("s")


@pytest.mark.skipif(
    PANDAS_VERSION >= (3,),
    reason="pandas 3.0+ disallows this kind of inplace modification",
)
@pytest.mark.skipif(
    PANDAS_VERSION < (1, 4),
    reason="pandas pre 1.4 doesn't change the type on inplace modification",
)
@pytest.mark.filterwarnings("ignore:Setting an item of incompatible")
def test_pandas_inplace_modification_1267() -> None:
    pytest.importorskip("pandas")
    import pandas as pd

    s = pd.Series([1, 2, 3])
    snw = nw.from_native(s, series_only=True)
    assert snw.dtype == nw.Int64
    s[0] = 999.5
    assert snw.dtype == nw.Float64


def test_pandas_fixed_offset_1302() -> None:
    pytest.importorskip("pyarrow")
    pytest.importorskip("pandas")
    import pandas as pd

    result = nw.from_native(
        pd.Series(pd.to_datetime(["2020-01-01T00:00:00.000000000+01:00"])),
        series_only=True,
    ).dtype
    expected_timezone = "UTC+01:00" if PANDAS_VERSION >= (2,) else "pytz.FixedOffset(60)"
    assert result == nw.Datetime("ns", expected_timezone)

    if PANDAS_VERSION >= (2,):
        result = nw.from_native(
            pd.Series(
                pd.to_datetime(["2020-01-01T00:00:00.000000000+01:00"])
            ).convert_dtypes(dtype_backend="pyarrow"),
            series_only=True,
        ).dtype
        assert result == nw.Datetime("ns", "+01:00")
    else:  # pragma: no cover
        pass


def from_native_collect_schema(native: IntoFrame) -> nw.Schema:
    return nw.from_native(native).collect_schema()


def test_huge_int() -> None:
    pytest.importorskip("duckdb")
    pytest.importorskip("polars")

    import duckdb
    import polars as pl

    df = pl.DataFrame({"a": [1, 2, 3]})

    if POLARS_VERSION >= (1, 18):
        result = from_native_collect_schema(df.select(pl.col("a").cast(pl.Int128)))
        assert result["a"] == nw.Int128
        assert result.to_polars()["a"] == pl.Int128
    else:  # pragma: no cover
        # Int128 was not available yet
        pass
    if POLARS_VERSION >= (1, 34):
        result = from_native_collect_schema(df.select(pl.col("a").cast(pl.UInt128)))
        assert result["a"] == nw.UInt128
        assert result.to_polars()["a"] == pl.UInt128
    else:  # pragma: no cover
        pass

    rel = duckdb.sql("""
        select cast(a as int128) as a
        from df
                     """)
    with pytest.warns(PerformanceWarning):
        result = nw.from_native(rel).schema
    assert result["a"] == nw.Int128

    rel = duckdb.sql("""
        select cast(a as uint128) as a
        from df
                     """)
    result = from_native_collect_schema(rel)
    assert result["a"] == nw.UInt128

    # TODO(unassigned): once other libraries support Int128/UInt128,
    # add tests for them too


@pytest.mark.skipif(PANDAS_VERSION < (1, 5), reason="too old for pyarrow")
def test_decimal() -> None:
    pytest.importorskip("duckdb")
    pytest.importorskip("polars")

    import duckdb
    import polars as pl

    df = pl.DataFrame({"a": [1]}, schema={"a": pl.Decimal})
    result = nw.from_native(df).schema
    assert result["a"] == nw.Decimal
    rel = duckdb.sql("""
        select *
        from df
                     """)
    result = nw.from_native(rel).collect_schema()
    assert result["a"] == nw.Decimal
    result = nw.from_native(
        df.to_pandas(use_pyarrow_extension_array=True)
    ).collect_schema()
    assert result["a"] == nw.Decimal
    result = nw.from_native(df.to_arrow()).collect_schema()
    assert result["a"] == nw.Decimal


def test_dtype_is_x() -> None:
    dtypes = (
        nw.Array,
        nw.Boolean,
        nw.Categorical,
        nw.Date,
        nw.Datetime,
        nw.Decimal,
        nw.Duration,
        nw.Enum,
        nw.Float32,
        nw.Float64,
        nw.Int8,
        nw.Int16,
        nw.Int32,
        nw.Int64,
        nw.Int128,
        nw.List,
        nw.Object,
        nw.String,
        nw.Struct,
        nw.Time,
        nw.UInt8,
        nw.UInt16,
        nw.UInt32,
        nw.UInt64,
        nw.UInt128,
        nw.Unknown,
        nw.Binary,
    )

    is_signed_integer = {nw.Int8, nw.Int16, nw.Int32, nw.Int64, nw.Int128}
    is_unsigned_integer = {nw.UInt8, nw.UInt16, nw.UInt32, nw.UInt64, nw.UInt128}
    is_float = {nw.Float32, nw.Float64}
    is_decimal = {nw.Decimal}
    is_temporal = {nw.Datetime, nw.Date, nw.Duration, nw.Time}
    is_nested = {nw.Array, nw.List, nw.Struct}
    is_boolean = {nw.Boolean}

    for dtype in dtypes:
        assert dtype.is_numeric() == (
            dtype
            in is_signed_integer.union(is_unsigned_integer)
            .union(is_float)
            .union(is_decimal)
        )
        assert dtype.is_integer() == (
            dtype in is_signed_integer.union(is_unsigned_integer)
        )
        assert dtype.is_signed_integer() == (dtype in is_signed_integer)
        assert dtype.is_unsigned_integer() == (dtype in is_unsigned_integer)
        assert dtype.is_float() == (dtype in is_float)
        assert dtype.is_decimal() == (dtype in is_decimal)
        assert dtype.is_temporal() == (dtype in is_temporal)
        assert dtype.is_nested() == (dtype in is_nested)
        assert dtype.is_boolean() == (dtype in is_boolean)


@pytest.mark.skipif(POLARS_VERSION < (1, 18), reason="too old for Int128")
def test_huge_int_to_native() -> None:
    pytest.importorskip("duckdb")
    pytest.importorskip("polars")

    import duckdb
    import polars as pl

    df = pl.DataFrame({"a": [1, 2, 3]})
    df_casted = (
        nw.from_native(df).with_columns(a_int=nw.col("a").cast(nw.Int128())).to_native()
    )
    assert df_casted.schema["a_int"] == pl.Int128

    rel = duckdb.sql("""
        select cast(a as int64) as a
        from df
                     """)
    result = (
        nw.from_native(rel)
        .with_columns(
            a_int=nw.col("a").cast(nw.Int128()), a_unit=nw.col("a").cast(nw.UInt128())
        )
        .select("a_int", "a_unit")
        .to_native()
    )
    type_a_int, type_a_unit = result.types
    assert type_a_int == "HUGEINT"
    assert type_a_unit == "UHUGEINT"


@pytest.mark.parametrize(
    ("precision", "scale"), [(None, 1), (None, 20), (2, 1), (10, 1), (10, 8)]
)
def test_cast_decimal_to_native(
    request: pytest.FixtureRequest,
    constructor: Constructor,
    precision: int | None,
    scale: int,
) -> None:
    if "dask" in str(constructor):
        request.applymarker(pytest.mark.xfail(reason="Unsupported dtype"))

    if "polars" in str(constructor) and POLARS_VERSION < (1, 0, 0):
        pytest.skip(reason="too old to convert to decimal")

    data = {"a": [1.1, 2.2, 3.3]}

    df = nw.from_native(constructor(data))

    if df.implementation.is_pandas_like() and (
        PYARROW_VERSION == (0, 0, 0) or PANDAS_VERSION < (2, 2)
    ):
        pytest.skip(reason="pyarrow is required to convert to decimal dtype")

    native_result = df.with_columns(
        a=nw.col("a").cast(nw.Decimal(precision=precision, scale=scale))
    ).to_native()

    schema = nw.from_native(native_result).collect_schema()
    assert schema["a"] == nw.Decimal(precision, scale)


@pytest.mark.parametrize(
    ("precision", "scale", "exception", "msg"),
    [
        (2.1, 0, TypeError, "precision must be a positive integer between 0 and 38"),
        ("foo", 0, TypeError, "precision must be a positive integer between 0 and 38"),
        (-1, 0, ValueError, "precision must be a positive integer between 0 and 38"),
        (39, 0, ValueError, "precision must be a positive integer between 0 and 38"),
        (None, 2.1, TypeError, "scale must be a positive integer"),
        (None, "foo", TypeError, "scale must be a positive integer"),
        (None, -1, ValueError, "scale must be a positive integer"),
        (2, 3, InvalidOperationError, "scale must be less than or equal to precision"),
    ],
)
def test_decimal_invalid(
    precision: int | None, scale: int, exception: type[Exception], msg: str
) -> None:
    with pytest.raises(exception, match=msg):
        nw.Decimal(precision=precision, scale=scale)


@pytest.mark.parametrize(
    "categories", [["a", "b"], enum.Enum("Test", "a b"), [1, 2, 3], []]
)
def test_enum_valid(categories: Iterable[Any] | type[enum.Enum]) -> None:
    dtype = nw.Enum(categories)
    assert dtype == nw.Enum
    assert len(dtype.categories) == len([*categories])


def test_enum_valid_numpy() -> None:
    pytest.importorskip("numpy")
    import numpy as np

    categories = [np.str_("a"), np.str_("b")]
    dtype = nw.Enum(categories)
    assert dtype == nw.Enum
    assert len(dtype.categories) == len([*categories])


def test_enum_from_series() -> None:
    pytest.importorskip("polars")
    import polars as pl

    elements = "a", "d", "e", "b", "c"
    categories = pl.Series(elements)
    categories_nw = nw.from_native(categories, series_only=True)
    assert nw.Enum(categories_nw).categories == elements
    assert nw.Enum(categories).categories == elements


def test_enum_categories_immutable() -> None:
    dtype = nw.Enum(["a", "b"])
    with pytest.raises(TypeError, match="does not support item assignment"):
        dtype.categories[0] = "c"  # type: ignore[index]
    with pytest.raises(AttributeError):
        dtype.categories = "a", "b", "c"  # type: ignore[misc]


def test_enum_repr_pd() -> None:
    pytest.importorskip("pandas")
    import pandas as pd

    df = nw.from_native(
        pd.DataFrame(
            {"a": ["broccoli", "cabbage"]}, dtype=pd.CategoricalDtype(ordered=True)
        )
    )
    dtype = df.schema["a"]
    assert isinstance(dtype, nw.Enum)
    assert dtype.categories == ("broccoli", "cabbage")
    assert "Enum(categories=['broccoli', 'cabbage'])" in str(dtype)


def test_enum_repr_pl() -> None:
    pytest.importorskip("polars")
    import polars as pl

    df = nw.from_native(
        pl.DataFrame(
            {"a": ["broccoli", "cabbage"]}, schema={"a": pl.Enum(["broccoli", "cabbage"])}
        )
    )
    dtype = df.schema["a"]
    assert isinstance(dtype, nw.Enum)
    assert dtype.categories == ("broccoli", "cabbage")
    assert "Enum(categories=['broccoli', 'cabbage'])" in repr(dtype)


def test_enum_repr() -> None:
    result = nw.Enum(["a", "b"])
    assert "Enum(categories=['a', 'b'])" in repr(result)
    result = nw.Enum(nw.Implementation)
    assert "Enum(categories=['pandas', 'modin', 'cudf'" in repr(result)


def test_enum_hash() -> None:
    dtype = nw.Enum(["a", "b"])
    assert_equal_hash(dtype, nw.Enum(["a", "b"]))
    with pytest.raises(AssertionError):
        assert_equal_hash(dtype, nw.Enum(["a", "b", "c"]))


@pytest.mark.xfail(
    reason="https://github.com/narwhals-dev/narwhals/pull/3213#discussion_r2437271987"
)
@pytest.mark.parametrize("dtype_name", ["Datetime", "Duration", "Enum"])
def test_dtype_repr_versioned(dtype_name: str) -> None:
    from narwhals.stable import v1 as nw_v1

    dtype_class_main = getattr(nw, dtype_name)
    dtype_class_v1 = getattr(nw_v1, dtype_name)

    assert dtype_class_main is not dtype_class_v1
    assert repr(dtype_class_main) != repr(dtype_class_v1)


def test_datetime_w_tz_duckdb() -> None:
    pytest.importorskip("duckdb")
    import duckdb

    duckdb.sql("""set timezone = 'Europe/Amsterdam'""")
    df = nw.from_native(
        duckdb.sql("""select * from values (timestamptz '2020-01-01')df(a)""")
    )
    result = df.collect_schema()
    assert result["a"] == nw.Datetime("us", "Europe/Amsterdam")
    duckdb.sql("""set timezone = 'Asia/Kathmandu'""")
    result = df.collect_schema()
    assert result["a"] == nw.Datetime("us", "Asia/Kathmandu")

    df = nw.from_native(
        duckdb.sql(
            """select * from values (timestamptz '2020-01-01', [[timestamptz '2020-01-02']])df(a,b)"""
        )
    )
    result = df.collect_schema()
    assert result["a"] == nw.Datetime("us", "Asia/Kathmandu")
    assert result["b"] == nw.List(nw.List(nw.Datetime("us", "Asia/Kathmandu")))


@pytest.mark.slow
def test_datetime_w_tz_pyspark() -> None:  # pragma: no cover
    pytest.importorskip("pyspark")
    session = pyspark_session()

    df = nw.from_native(
        session.createDataFrame([(datetime(2020, 1, 1, tzinfo=timezone.utc),)], ["a"])
    )
    result = df.collect_schema()
    assert result["a"] == nw.Datetime("us", "UTC")
    df = nw.from_native(
        session.createDataFrame([([datetime(2020, 1, 1, tzinfo=timezone.utc)],)], ["a"])
    )
    result = df.collect_schema()
    assert result["a"] == nw.List(nw.Datetime("us", "UTC"))


def test_dtype_base_type_non_nested(non_nested_type: type[NonNestedDType]) -> None:
    assert non_nested_type.base_type() is non_nested_type().base_type()


def test_dtype_base_type_nested(nested_dtype: NestedOrEnumDType) -> None:
    base = nested_dtype.base_type()
    assert base.base_type() == nested_dtype.base_type()


@pytest.mark.parametrize(
    ("dtype", "context"),
    [
        (nw.Datetime("ns"), does_not_warn()),
        (nw.Datetime, does_not_warn()),
        (nw.Datetime(), pytest.warns(UserWarning, match="time_unit")),
        (nw.Datetime("us"), pytest.warns(UserWarning, match="time_unit='us'")),
        (nw.Datetime("s"), pytest.warns(UserWarning, match="time_unit='s'")),
    ],
)
def test_pandas_datetime_ignored_time_unit_warns(
    constructor_pandas_like: ConstructorPandasLike,
    dtype: nw.Datetime | type[nw.Datetime],
    context: AbstractContextManager[Any],
) -> None:
    data = {"a": [datetime(2001, 1, 1), None, datetime(2001, 1, 3)]}
    expr = nw.col("a").cast(dtype)
    df = nw.from_native(constructor_pandas_like(data))
    ctx = does_not_warn() if PANDAS_VERSION >= (2,) else context
    with ctx:
        df.select(expr)


def test_dtype___slots___non_nested(non_nested_type: type[NonNestedDType]) -> None:
    dtype = non_nested_type()
    with pytest.raises(AttributeError):
        dtype.i_dont_exist = 100  # type: ignore[union-attr]
    with pytest.raises(AttributeError):
        dtype.__dict__  # noqa: B018
    _ = dtype.__slots__


def test_dtype___slots___nested(nested_dtype: NestedOrEnumDType) -> None:
    with pytest.raises(AttributeError):
        nested_dtype.i_dont_exist = 999  # type: ignore[union-attr]
    with pytest.raises(AttributeError):
        nested_dtype.__dict__  # noqa: B018
    slots = nested_dtype.__slots__
    assert len(slots) != 0, slots



from __future__ import annotations

from typing import TYPE_CHECKING

import nox

if TYPE_CHECKING:
    from nox.sessions import Session

nox.options.default_venv_backend = "uv"
nox.options.reuse_venv = "yes"

PYTHON_VERSIONS = ["3.8", "3.9", "3.10", "3.11", "3.12"]


def run_common(session: Session, coverage_threshold: float) -> None:
    if session.python == "3.8":
        session.install("-e . --group dev-core")
    elif session.python == "3.12":
        session.install("-e .[dask,modin] --group dev-core --group extra")
    else:
        session.install("-e .[dask,modin,pyspark,ibis] --group dev-core --group extra")

    session.run(
        "pytest",
        "tests",
        "--cov=narwhals",
        "--cov=tests",
        f"--cov-fail-under={coverage_threshold}",
        "--runslow",
    )

    if session.python == "3.12":
        session.run("pytest", "narwhals", "--doctest-modules")


@nox.session(python=PYTHON_VERSIONS)  # type: ignore[misc]
def pytest_coverage(session: Session) -> None:
    coverage_threshold = 85 if session.python == "3.8" else 100

    run_common(session, coverage_threshold)


@nox.session(python=PYTHON_VERSIONS[0])  # type: ignore[misc]
@nox.parametrize("pandas_version", ["0.25.3", "1.1.5"])  # type: ignore[misc]
def min_and_old_versions(session: Session, pandas_version: str) -> None:
    session.install(
        f"pandas=={pandas_version}",
        "polars==0.20.4",
        "numpy==1.17.5",
        "pyarrow==11.0.0",
        "scipy==1.5.0",
        "scikit-learn==1.1.0",
        "tzdata",
    )
    if pandas_version == "1.1.5":
        session.install("pyspark==3.3.0")
    run_common(session, coverage_threshold=50)


@nox.session(python=PYTHON_VERSIONS[-1])  # type: ignore[misc]
def nightly_versions(session: Session) -> None:
    session.install("polars")

    session.install(  # pandas nightly
        "--pre",
        "--extra-index-url",
        "https://pypi.anaconda.org/scientific-python-nightly-wheels/simple",
        "pandas",
    )

    session.install(  # numpy nightly
        "--pre",
        "--extra-index-url",
        "https://pypi.anaconda.org/scientific-python-nightly-wheels/simple",
        "numpy",
    )

    session.run("uv", "pip", "install", "pip")
    session.run(  # dask nightly
        "pip",
        "install",
        "git+https://github.com/dask/distributed",
        "git+https://github.com/dask/dask",
        "git+https://github.com/dask/dask-expr",
    )
    run_common(session, coverage_threshold=50)



from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, Any

import pytest

import narwhals as nw
from tests.utils import ConstructorEager, assert_equal_data, assert_equal_series

if TYPE_CHECKING:
    from collections.abc import Collection


def series(frame: ConstructorEager, name: str, values: Collection[Any]) -> nw.Series[Any]:
    return nw.from_native(frame({name: values})).get_column(name)


@pytest.mark.filterwarnings(
    "ignore:.*all arguments of to_dict except for the argument:FutureWarning"
)
@pytest.mark.parametrize(
    ("data", "indices", "values", "expected"),
    [
        ([142, 124, 13], [0, 2, 1], (142, 124, 13), [142, 13, 124]),
        ([1, 2, 3], 0, 999, [999, 2, 3]),
        (
            [16, 12, 10, 9, 6, 5, 2],
            (6, 1, 0, 5, 3, 2, 4),
            [16, 12, 10, 9, 6, 5, 2],
            [10, 12, 5, 6, 2, 9, 16],
        ),
        ([5.5, 9.2, 1.0], (), (), [5.5, 9.2, 1.0]),
    ],
    ids=["single-series", "integer", "unordered-indices", "empty-indices"],
)
def test_scatter(
    data: list[Any],
    indices: int | Collection[int],
    values: int | Collection[int],
    expected: list[Any],
    constructor_eager: ConstructorEager,
) -> None:
    constructor = partial(series, constructor_eager)
    s = constructor("s", data)
    df = s.to_frame().with_row_index("dont change me")
    unchanged_indexed = df.to_dict(as_series=False)
    assert_equal_series(s.scatter(indices, values), expected, "s")
    if not isinstance(indices, int):
        assert_equal_series(s.scatter(constructor("i", indices), values), expected, "s")
    if not isinstance(values, int):
        assert_equal_series(s.scatter(indices, constructor("v", values)), expected, "s")
    assert_equal_data(df, unchanged_indexed)


def test_scatter_pandas_index() -> None:
    pytest.importorskip("pandas")
    import pandas as pd

    s = nw.from_native(pd.Series([2, 3, 6], index=[1, 0, 2]), series_only=True)
    result = s.scatter([1, 0, 2], s)
    expected = pd.Series([3, 2, 6], index=[1, 0, 2])
    pd.testing.assert_series_equal(result.to_native(), expected)


def test_scatter_2862(constructor_eager: ConstructorEager) -> None:
    s = series(constructor_eager, "a", [1, 2, 3])
    assert_equal_series(s.scatter(1, 999), [1, 999, 3], "a")
    assert_equal_series(s.scatter([0, 2], [999, 888]), [999, 2, 888], "a")
    assert_equal_series(s.scatter([2, 0], [999, 888]), [888, 2, 999], "a")



from __future__ import annotations

import pytest

import narwhals as nw
from tests.utils import DUCKDB_VERSION, Constructor, assert_equal_data


def test_sumh_broadcasting(constructor: Constructor) -> None:
    if "duckdb" in str(constructor) and DUCKDB_VERSION < (1, 3):
        pytest.skip()
    data = {"a": [1, 2, 3], "b": [4, 5, 6], "i": [0, 1, 2]}
    df = nw.from_native(constructor(data))
    result = df.with_columns(
        any=nw.any_horizontal(nw.sum("a", "b").cast(nw.Boolean), ignore_nulls=True),
        all=nw.all_horizontal(nw.sum("a", "b").cast(nw.Boolean), ignore_nulls=True),
        max=nw.max_horizontal(nw.sum("a"), nw.sum("b")),
        min=nw.min_horizontal(nw.sum("a"), nw.sum("b")),
        sum=nw.sum_horizontal(nw.sum("a"), nw.sum("b")),
        mean=nw.mean_horizontal(nw.sum("a"), nw.sum("b")),
    ).sort("i")
    expected = {
        "a": [1, 2, 3],
        "b": [4, 5, 6],
        "i": [0, 1, 2],
        "any": [True, True, True],
        "all": [True, True, True],
        "max": [15, 15, 15],
        "min": [6, 6, 6],
        "sum": [21, 21, 21],
        "mean": [10.5, 10.5, 10.5],
    }
    assert_equal_data(result, expected)



from __future__ import annotations

import re
from contextlib import AbstractContextManager, nullcontext as does_not_raise
from typing import TYPE_CHECKING, Any, Callable

import pytest

import narwhals as nw
from narwhals.testing import assert_series_equal
from tests.utils import PANDAS_VERSION, POLARS_VERSION, PYARROW_VERSION

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

    from narwhals.typing import IntoSchema, IntoSeriesT
    from tests.conftest import Data
    from tests.utils import ConstructorEager

    SetupFn: TypeAlias = Callable[[nw.Series[Any]], tuple[nw.Series[Any], nw.Series[Any]]]


def _assertion_error(detail: str) -> pytest.RaisesExc:
    return pytest.raises(
        AssertionError, match=re.escape(f"Series are different ({detail})")
    )


def series_from_native(native: IntoSeriesT) -> nw.Series[IntoSeriesT]:
    return nw.from_native(native, series_only=True)


def test_self_equal(
    request: pytest.FixtureRequest,
    constructor_eager: ConstructorEager,
    testing_data: Data,
    testing_schema: IntoSchema,
) -> None:
    """Test that a series is equal to itself, including nested dtypes with nulls."""
    if "cudf" in str(constructor_eager):
        # TODO(FBruzzesi): Investigate which conversion is failing aside from nested dtypes
        request.applymarker(pytest.mark.xfail)
    if "pandas" in str(constructor_eager):
        if PANDAS_VERSION < (2, 2):  # pragma: no cover
            reason = "Pandas too old for nested dtypes"
            pytest.skip(reason=reason)
        pytest.importorskip("pyarrow")

    if "pyarrow_table" in str(constructor_eager) and PYARROW_VERSION < (
        15,
        0,
    ):  # pragma: no cover
        reason = (
            "pyarrow.lib.ArrowNotImplementedError: Unsupported cast from string to "
            "dictionary using function cast_dictionary"
        )
        pytest.skip(reason=reason)

    if "pyarrow_table" in str(constructor_eager):
        # Replace Enum with Categorical, since Pyarrow does not support Enum
        data = dict(testing_data)
        schema = {**testing_schema, "enum": nw.Categorical()}
    else:
        data = dict(testing_data)
        schema = dict(testing_schema)

    df = nw.from_native(constructor_eager(data), eager_only=True)
    for name, dtype in schema.items():
        assert_series_equal(df[name].cast(dtype), df[name].cast(dtype))


def test_implementation_mismatch() -> None:
    """Test that different implementations raise an error."""
    pytest.importorskip("pandas")
    pytest.importorskip("pyarrow")

    import pandas as pd
    import pyarrow as pa

    with _assertion_error("implementation mismatch"):
        assert_series_equal(
            series_from_native(pd.Series([1])),
            series_from_native(pa.chunked_array([[2]])),  # type: ignore[misc] # pyright: ignore[reportArgumentType]
        )


@pytest.mark.parametrize(
    ("setup_fn", "error_msg"),
    [
        (lambda s: (s, s.head(2)), "length mismatch"),
        (lambda s: (s.cast(nw.UInt32()), s.cast(nw.Int64())), "dtype mismatch"),
        (lambda s: (s.rename("foo"), s.rename("bar")), "name mismatch"),
    ],
)
def test_metadata_checks(
    constructor_eager: ConstructorEager, setup_fn: SetupFn, error_msg: str
) -> None:
    """Test metadata validation (length, dtype, name)."""
    series = nw.from_native(constructor_eager({"a": [1, 2, 3]}), eager_only=True)["a"]
    left, right = setup_fn(series)

    with _assertion_error(error_msg):
        assert_series_equal(left, right)


@pytest.mark.parametrize(
    ("setup_fn", "error_msg", "check_dtypes", "check_names"),
    [
        (lambda s: (s, s.cast(nw.UInt32())), "dtype mismatch", True, False),
        (lambda s: (s, s.cast(nw.UInt32()).rename("baz")), "dtype mismatch", True, True),
        (lambda s: (s, s.rename("baz")), "name mismatch", False, True),
    ],
)
def test_metadata_checks_with_flags(
    constructor_eager: ConstructorEager,
    setup_fn: SetupFn,
    error_msg: str,
    *,
    check_dtypes: bool,
    check_names: bool,
) -> None:
    """Test the effect of check_dtypes and check_names flags."""
    series = nw.from_native(constructor_eager({"a": [1, 2, 3]}), eager_only=True)["a"]
    left, right = setup_fn(series)

    with _assertion_error(error_msg):
        assert_series_equal(
            left, right, check_dtypes=check_dtypes, check_names=check_names
        )

    assert_series_equal(left, right, check_dtypes=False, check_names=False)


@pytest.mark.parametrize(
    ("dtype", "check_order", "context"),
    [
        (nw.List(nw.Int32()), False, pytest.raises(NotImplementedError)),
        (nw.List(nw.Int32()), True, does_not_raise()),
        (nw.Int32(), False, does_not_raise()),
        (nw.Int32(), True, does_not_raise()),
    ],
)
def test_check_order(
    request: pytest.FixtureRequest,
    constructor_eager: ConstructorEager,
    dtype: nw.dtypes.DType,
    *,
    check_order: bool,
    context: AbstractContextManager[Any],
) -> None:
    """Test check_order behavior with nested and simple data."""
    if "cudf" in str(constructor_eager) and dtype.is_nested():
        reason = "NotImplementedError"
        request.applymarker(pytest.mark.xfail(reason=reason))

    if "pandas" in str(constructor_eager):
        if PANDAS_VERSION < (2, 2) and dtype.is_nested():  # pragma: no cover
            reason = "Pandas too old for nested dtypes"
            pytest.skip(reason=reason)
        pytest.importorskip("pyarrow")

    data: list[Any] = [[1, 2, 3]] if dtype.is_nested() else [1, 2, 3]
    frame = nw.from_native(constructor_eager({"a": data}), eager_only=True)
    left = right = frame["a"].cast(dtype)

    with context:
        assert_series_equal(left, right, check_order=check_order, check_names=False)


@pytest.mark.parametrize(
    "null_data",
    [
        {"left": ["x", "y", None], "right": ["x", None, "y"]},  # Different null position
        {"left": ["x", None, None], "right": [None, "x", "y"]},  # Different null counts
    ],
)
def test_null_mismatch(constructor_eager: ConstructorEager, null_data: Data) -> None:
    """Test null value mismatch detection."""
    frame = nw.from_native(constructor_eager(null_data), eager_only=True)
    left, right = frame["left"], frame["right"]
    with _assertion_error("null value mismatch"):
        assert_series_equal(left, right, check_names=False)


@pytest.mark.parametrize(
    ("check_exact", "abs_tol", "rel_tol", "context"),
    [
        (True, 1e-3, 1e-3, _assertion_error("exact value mismatch")),
        (False, 1e-3, 1e-3, _assertion_error("values not within tolerance")),
        (False, 2e-1, 2e-1, does_not_raise()),
    ],
)
def test_numeric(
    constructor_eager: ConstructorEager,
    *,
    check_exact: bool,
    abs_tol: float,
    rel_tol: float,
    context: AbstractContextManager[Any],
) -> None:
    data = {
        "left": [1.0, float("nan"), float("inf"), None, 1.1],
        "right": [1.01, float("nan"), float("inf"), None, 1.11],
    }

    frame = nw.from_native(constructor_eager(data), eager_only=True)
    left, right = frame["left"], frame["right"]
    with context:
        assert_series_equal(
            left,
            right,
            check_names=False,
            check_exact=check_exact,
            abs_tol=abs_tol,
            rel_tol=rel_tol,
        )


@pytest.mark.parametrize(
    ("l_vals", "r_vals", "check_exact", "context", "dtype"),
    [
        (
            [["foo", "bar"]],
            [["foo", None]],
            True,
            _assertion_error("nested value mismatch"),
            nw.List(nw.String()),
        ),
        (
            [["foo", "bar"]],
            [["foo", None]],
            True,
            _assertion_error("nested value mismatch"),
            nw.Array(nw.String(), 2),
        ),
        (
            [[0.0, 0.1]],
            [[0.1, 0.1]],
            True,
            _assertion_error("nested value mismatch"),
            nw.List(nw.Float32()),
        ),
        (
            [[0.0, 0.1]],
            [[0.1, 0.1]],
            True,
            _assertion_error("nested value mismatch"),
            nw.Array(nw.Float32(), 2),
        ),
        ([[0.0, 1e-10]], [[1e-10, 0.0]], False, does_not_raise(), nw.List(nw.Float64())),
        (
            [[0.0, 1e-10]],
            [[1e-10, 0.0]],
            False,
            does_not_raise(),
            nw.Array(nw.Float64(), 2),
        ),
    ],
)
def test_list_like(
    request: pytest.FixtureRequest,
    constructor_eager: ConstructorEager,
    l_vals: list[list[Any]],
    r_vals: list[list[Any]],
    *,
    check_exact: bool,
    context: AbstractContextManager[Any],
    dtype: nw.dtypes.DType,
) -> None:
    if "cudf" in str(constructor_eager):
        reason = "NotImplementedError"
        request.applymarker(pytest.mark.xfail(reason=reason))

    if "pandas" in str(constructor_eager):
        if PANDAS_VERSION < (2, 2):  # pragma: no cover
            reason = "Pandas too old for nested dtypes"
            pytest.skip(reason=reason)
        pytest.importorskip("pyarrow")

    if (
        "pyarrow_table" in str(constructor_eager)
        and PYARROW_VERSION < (14, 0)
        and dtype == nw.Array
    ):  # pragma: no cover
        reason = (
            "pyarrow.lib.ArrowNotImplementedError: Unsupported cast from "
            "list<item: string> to fixed_size_list using function cast_fixed_size_list"
        )
        pytest.skip(reason=reason)

    data = {"left": l_vals, "right": r_vals}
    frame = nw.from_native(constructor_eager(data), eager_only=True)
    left, right = frame["left"].cast(dtype), frame["right"].cast(dtype)
    with context:
        assert_series_equal(left, right, check_names=False, check_exact=check_exact)


@pytest.mark.parametrize(
    ("l_vals", "r_vals", "check_exact", "context"),
    [
        (
            [{"a": 0.0, "b": ["orca"]}, None],
            [{"a": 1e-10, "b": ["orca"]}, None],
            True,
            _assertion_error("exact value mismatch"),
        ),
        (
            [{"a": 0.0, "b": ["beluga"]}, None],
            [{"a": 0.0, "b": ["orca"]}, None],
            False,
            _assertion_error("exact value mismatch"),
        ),
        (
            [{"a": 0.0, "b": ["orca"]}, None],
            [{"a": 1e-10, "b": ["orca"]}, None],
            False,
            does_not_raise(),
        ),
    ],
)
def test_struct(
    request: pytest.FixtureRequest,
    constructor_eager: ConstructorEager,
    l_vals: list[dict[str, Any]],
    r_vals: list[dict[str, Any]],
    *,
    check_exact: bool,
    context: AbstractContextManager[Any],
) -> None:
    if "cudf" in str(constructor_eager):
        reason = "NotImplementedError"
        request.applymarker(pytest.mark.xfail(reason=reason))

    if "pandas" in str(constructor_eager):
        if PANDAS_VERSION < (2, 2):  # pragma: no cover
            reason = "Pandas too old for nested dtypes"
            pytest.skip(reason=reason)
        pytest.importorskip("pyarrow")

    dtype = nw.Struct({"a": nw.Float32(), "b": nw.List(nw.String())})
    data = {"left": l_vals, "right": r_vals}
    frame = nw.from_native(constructor_eager(data), eager_only=True)
    left, right = frame["left"].cast(dtype), frame["right"].cast(dtype)
    with context:
        assert_series_equal(left, right, check_names=False, check_exact=check_exact)


def test_non_nw_series() -> None:
    pytest.importorskip("pandas")

    import pandas as pd

    with pytest.raises(
        TypeError, match=re.escape("Expected `narwhals.Series` instance, found")
    ):
        assert_series_equal(
            left=pd.Series([1]),  # type: ignore[arg-type]
            right=pd.Series([2]),  # type: ignore[arg-type]
        )


@pytest.mark.parametrize(
    ("categorical_as_str", "context"),
    [
        (True, does_not_raise()),
        (
            False,
            pytest.raises(
                AssertionError,
                match="Cannot compare categoricals coming from different sources",
            ),
        ),
    ],
)
def test_categorical_as_str(
    request: pytest.FixtureRequest,
    constructor_eager: ConstructorEager,
    *,
    categorical_as_str: bool,
    context: AbstractContextManager[Any],
) -> None:
    if (
        "polars" in str(constructor_eager)
        and POLARS_VERSION >= (1, 32)
        and not categorical_as_str
    ):
        # https://github.com/pola-rs/polars/pull/23016 removed StringCache, it still
        # exists but it does nothing in python.
        request.applymarker(pytest.mark.xfail)

    if "pyarrow_table" in str(constructor_eager) and not categorical_as_str:
        # pyarrow dictionary dtype compares values, not the encoding.
        request.applymarker(pytest.mark.xfail)

    if "pyarrow_table" in str(constructor_eager) and PYARROW_VERSION < (
        15,
        0,
    ):  # pragma: no cover
        reason = (
            "pyarrow.lib.ArrowNotImplementedError: Unsupported cast from string to "
            "dictionary using function cast_dictionary"
        )
        pytest.skip(reason=reason)

    data = {
        "left": ["beluga", "dolphin", "narwhal", "orca"],
        "right": ["unicorn", "orca", "narwhal", "orca"],
    }
    frame = nw.from_native(constructor_eager(data), eager_only=True)
    left = frame["left"].cast(nw.Categorical())[2:]
    right = frame["right"].cast(nw.Categorical())[2:]

    with context:
        assert_series_equal(
            left, right, check_names=False, categorical_as_str=categorical_as_str
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

import ibis
from ibis.expr.datatypes import Timestamp

from narwhals._sql.expr_str import SQLExprStringNamespace
from narwhals._utils import _is_naive_format, not_implemented

if TYPE_CHECKING:
    import ibis.expr.types as ir
    from typing_extensions import TypeAlias

    from narwhals._ibis.expr import IbisExpr

IntoStringValue: TypeAlias = "str | ir.StringValue"


class IbisExprStringNamespace(SQLExprStringNamespace["IbisExpr"]):
    def strip_chars(self, characters: str | None) -> IbisExpr:
        if characters is not None:
            msg = "Ibis does not support `characters` argument in `str.strip_chars`"
            raise NotImplementedError(msg)

        return self.compliant._with_callable(lambda expr: expr.strip())

    def _replace_all(
        self, pattern: IntoStringValue, value: IntoStringValue
    ) -> Callable[..., ir.StringValue]:
        def fn(expr: ir.StringColumn) -> ir.StringValue:
            return expr.re_replace(pattern, value)

        return fn

    def _replace_all_literal(
        self, pattern: IntoStringValue, value: IntoStringValue
    ) -> Callable[..., ir.StringValue]:
        def fn(expr: ir.StringColumn) -> ir.StringValue:
            return expr.replace(pattern, value)  # pyright: ignore[reportArgumentType]

        return fn

    def replace_all(self, value: IbisExpr, pattern: str, *, literal: bool) -> IbisExpr:
        fn = self._replace_all_literal if literal else self._replace_all
        return self.compliant._with_elementwise(
            lambda expr, value: fn(pattern, value)(expr), value=value
        )

    def _to_datetime(self, format: str) -> Callable[..., ir.TimestampValue]:
        def fn(expr: ir.StringColumn) -> ir.TimestampValue:
            return expr.as_timestamp(format)

        return fn

    def _to_datetime_naive(self, format: str) -> Callable[..., ir.TimestampValue]:
        def fn(expr: ir.StringColumn) -> ir.TimestampValue:
            dtype: Any = Timestamp(timezone=None)
            return expr.as_timestamp(format).cast(dtype)

        return fn

    def to_datetime(self, format: str | None) -> IbisExpr:
        if format is None:
            msg = "Cannot infer format with Ibis backend"
            raise NotImplementedError(msg)
        fn = self._to_datetime_naive if _is_naive_format(format) else self._to_datetime
        return self.compliant._with_callable(fn(format))

    def to_date(self, format: str | None) -> IbisExpr:
        if format is None:
            msg = "Cannot infer format with Ibis backend"
            raise NotImplementedError(msg)

        def fn(expr: ir.StringColumn) -> ir.DateValue:
            return expr.as_date(format)

        return self.compliant._with_callable(fn)

    def pad_start(self, length: int, fill_char: str) -> IbisExpr:
        def _pad_start(expr: ir.StringColumn) -> ir.Value:
            padded = expr.lpad(length, fill_char)
            return ibis.cases((expr.notnull(), padded))

        return self.compliant._with_callable(_pad_start)

    def pad_end(self, length: int, fill_char: str) -> IbisExpr:
        def _pad_end(expr: ir.StringColumn) -> ir.Value:
            padded = expr.rpad(length, fill_char)
            return ibis.cases((expr.notnull(), padded))

        return self.compliant._with_callable(_pad_end)

    replace = not_implemented()
    to_titlecase = not_implemented()



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest

import narwhals as nw
import narwhals.stable.v1 as nw_v1

if TYPE_CHECKING:
    from collections.abc import Mapping

data: Mapping[str, Any] = {"a": [1, 2, 3], "b": [4.0, 5.0, 6.1], "z": ["x", "y", "z"]}


class InterchangeDataFrame:
    def __init__(self, df: CustomDataFrame) -> None:
        self._df = df

    def __dataframe__(self) -> InterchangeDataFrame:  # pragma: no cover
        return self

    def column_names(self) -> list[str]:
        return list(self._df._data.keys())

    def select_columns_by_name(self, columns: list[str]) -> InterchangeDataFrame:
        return InterchangeDataFrame(
            CustomDataFrame(
                {key: value for key, value in self._df._data.items() if key in columns}
            )
        )


class CustomDataFrame:
    def __init__(self, data: dict[str, Any]) -> None:
        self._data = data

    def __dataframe__(self, *, allow_copy: bool = True) -> InterchangeDataFrame:
        return InterchangeDataFrame(self)


def test_interchange() -> None:
    df = CustomDataFrame({"a": [1, 2, 3], "b": [4, 5, 6], "z": [1, 4, 2]})
    result = nw_v1.from_native(df, eager_or_interchange_only=True).select("a", "z")
    assert result.columns == ["a", "z"]


def test_interchange_non_v1() -> None:
    df = CustomDataFrame({"a": [1, 2, 3], "b": [4, 5, 6], "z": [1, 4, 2]})
    with pytest.raises(TypeError, match="Unsupported dataframe type"):
        nw.from_native(df)  # type: ignore[call-overload]


def test_interchange_ibis(
    tmpdir: pytest.TempdirFactory, request: pytest.FixtureRequest
) -> None:  # pragma: no cover
    pytest.importorskip("polars")
    pytest.importorskip("ibis")

    import ibis
    import polars as pl

    try:
        ibis.set_backend("duckdb")
    except ImportError:
        request.applymarker(pytest.mark.xfail)
    df_pl = pl.DataFrame(data)

    filepath = str(tmpdir / "file.parquet")  # type: ignore[operator]
    df_pl.write_parquet(filepath)

    tbl = ibis.read_parquet(filepath)
    df = nw_v1.from_native(tbl, eager_or_interchange_only=True)

    out_cols = df.select("a", "z").schema.names()

    assert out_cols == ["a", "z"]


def test_interchange_duckdb() -> None:
    pytest.importorskip("polars")
    pytest.importorskip("duckdb")

    import duckdb
    import polars as pl

    df_pl = pl.DataFrame(data)  # noqa: F841
    rel = duckdb.sql("select * from df_pl")
    df = nw_v1.from_native(rel, eager_or_interchange_only=True)

    out_cols = df.select("a", "z").schema.names()

    assert out_cols == ["a", "z"]



from __future__ import annotations

import narwhals as nw
from tests.utils import Constructor, ConstructorEager, assert_equal_data

data = {"a": [1.0, None, None, 3.0], "b": [1.0, None, 4.0, 5.0]}


def test_null_count_expr(constructor: Constructor) -> None:
    df = nw.from_native(constructor(data))
    result = df.select(nw.col("a", "b").null_count())
    expected = {"a": [2], "b": [1]}
    assert_equal_data(result, expected)


def test_null_count_series(constructor_eager: ConstructorEager) -> None:
    data = [1, 2, None]
    series = nw.from_native(constructor_eager({"a": data}), eager_only=True)["a"]
    result = series.null_count()
    assert result == 1



from __future__ import annotations

import operator
from typing import TYPE_CHECKING

from narwhals._compliant import LazyExprNamespace
from narwhals._compliant.any_namespace import ListNamespace

if TYPE_CHECKING:
    from sqlframe.base.column import Column

    from narwhals._spark_like.expr import SparkLikeExpr
    from narwhals.typing import NonNestedLiteral


class SparkLikeExprListNamespace(
    LazyExprNamespace["SparkLikeExpr"], ListNamespace["SparkLikeExpr"]
):
    def len(self) -> SparkLikeExpr:
        return self.compliant._with_elementwise(self.compliant._F.array_size)

    def unique(self) -> SparkLikeExpr:
        return self.compliant._with_elementwise(self.compliant._F.array_distinct)

    def contains(self, item: NonNestedLiteral) -> SparkLikeExpr:
        def func(expr: Column) -> Column:
            F = self.compliant._F
            return F.array_contains(expr, F.lit(item))

        return self.compliant._with_elementwise(func)

    def get(self, index: int) -> SparkLikeExpr:
        def _get(expr: Column) -> Column:
            return expr.getItem(index)

        return self.compliant._with_elementwise(_get)

    def min(self) -> SparkLikeExpr:
        def func(expr: Column) -> Column:
            F = self.compliant._F
            return F.array_min(expr)

        return self.compliant._with_elementwise(func)

    def max(self) -> SparkLikeExpr:
        def func(expr: Column) -> Column:
            F = self.compliant._F
            return F.array_max(F.array_compact(expr))

        return self.compliant._with_elementwise(func)

    def sum(self) -> SparkLikeExpr:
        def func(expr: Column) -> Column:
            F = self.compliant._F
            drop_nulls = F.array_compact(expr)
            len = F.array_size(drop_nulls)
            sum = F.aggregate(drop_nulls, F.lit(0.0), operator.add)
            return F.when((len.isNotNull()) & (len == 0), F.lit(0)).otherwise(sum)

        return self.compliant._with_elementwise(func)

    def mean(self) -> SparkLikeExpr:
        def func(expr: Column) -> Column:
            F = self.compliant._F
            return F.try_divide(
                F.aggregate(F.array_compact(expr), F.lit(0.0), operator.add),
                F.array_size(F.array_compact(expr)),
            )

        return self.compliant._with_elementwise(func)

    def median(self) -> SparkLikeExpr:
        def func(expr: Column) -> Column:
            F = self.compliant._F
            sorted_expr = F.array_compact(F.sort_array(expr))
            size = F.array_size(sorted_expr)
            mid_index = (size / 2).cast("int")
            impl = self.compliant._implementation
            if impl.is_sqlframe() and impl._backend_version() < (3, 44, 1):
                # indexing is different in sqlframe
                # https://github.com/eakmanrq/sqlframe/issues/568
                mid_index = mid_index + 1  # pragma: no cover
            odd_case = sorted_expr[mid_index]
            even_case = (sorted_expr[mid_index - 1] + sorted_expr[mid_index]) / 2
            return (
                F.when((size.isNull()) | (size == 0), F.lit(None))
                .when(size % 2 == 1, odd_case)
                .otherwise(even_case)
            )

        return self.compliant._with_elementwise(func)

    def sort(self, *, descending: bool, nulls_last: bool) -> SparkLikeExpr:
        def func(expr: Column) -> Column:
            F = self.compliant._F
            if not descending and nulls_last:
                return F.array_sort(expr)
            if descending and not nulls_last:
                impl = self.compliant._implementation
                rev = F.array_reverse if impl.is_sqlframe() else F.reverse
                return rev(F.array_sort(expr))
            return F.sort_array(expr, asc=not descending)

        return self.compliant._with_elementwise(func)



from __future__ import annotations

from functools import lru_cache, partial
from typing import TYPE_CHECKING, Any, Literal, cast, overload

import ibis
import ibis.expr.datatypes as ibis_dtypes

from narwhals._utils import Version, isinstance_or_issubclass

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping
    from datetime import timedelta

    import ibis.expr.types as ir
    from ibis.common.temporal import TimestampUnit
    from ibis.expr.datatypes import DataType as IbisDataType
    from typing_extensions import TypeAlias, TypeIs

    from narwhals._duration import IntervalUnit
    from narwhals._ibis.dataframe import IbisLazyFrame
    from narwhals._ibis.expr import IbisExpr
    from narwhals.dtypes import DType
    from narwhals.typing import IntoDType, PythonLiteral

IntoColumn: TypeAlias = "str | ir.Value | ir.Column"
SortFn: TypeAlias = "Callable[[IntoColumn], ir.Column]"
Incomplete: TypeAlias = Any
"""Marker for upstream issues."""


@overload
def lit(value: bool, dtype: None = ...) -> ir.BooleanScalar: ...  # noqa: FBT001
@overload
def lit(value: int, dtype: None = ...) -> ir.IntegerScalar: ...
@overload
def lit(value: float, dtype: None = ...) -> ir.FloatingScalar: ...
@overload
def lit(value: str, dtype: None = ...) -> ir.StringScalar: ...
@overload
def lit(value: PythonLiteral | ir.Value, dtype: None = ...) -> ir.Scalar: ...
@overload
def lit(value: Any, dtype: Any) -> Incomplete: ...
def lit(value: Any, dtype: Any | None = None) -> Incomplete:
    """Alias for `ibis.literal`."""
    literal: Incomplete = ibis.literal
    return literal(value, dtype)


asc_nulls_first = cast("SortFn", partial(ibis.asc, nulls_first=True))
asc_nulls_last = cast("SortFn", partial(ibis.asc, nulls_first=False))
desc_nulls_first = cast("SortFn", partial(ibis.desc, nulls_first=True))
desc_nulls_last = cast("SortFn", partial(ibis.desc, nulls_first=False))


BucketUnit: TypeAlias = Literal[
    "years",
    "quarters",
    "months",
    "days",
    "hours",
    "minutes",
    "seconds",
    "milliseconds",
    "microseconds",
    "nanoseconds",
]
TruncateUnit: TypeAlias = Literal[
    "Y", "Q", "M", "W", "D", "h", "m", "s", "ms", "us", "ns"
]

UNITS_DICT_BUCKET: Mapping[IntervalUnit, BucketUnit] = {
    "y": "years",
    "q": "quarters",
    "mo": "months",
    "d": "days",
    "h": "hours",
    "m": "minutes",
    "s": "seconds",
    "ms": "milliseconds",
    "us": "microseconds",
    "ns": "nanoseconds",
}

UNITS_DICT_TRUNCATE: Mapping[IntervalUnit, TruncateUnit] = {
    "y": "Y",
    "q": "Q",
    "mo": "M",
    "d": "D",
    "h": "h",
    "m": "m",
    "s": "s",
    "ms": "ms",
    "us": "us",
    "ns": "ns",
}

FUNCTION_REMAPPING = {
    "starts_with": "startswith",
    "ends_with": "endswith",
    "regexp_matches": "re_search",
    "str_split": "split",
    "dayofyear": "day_of_year",
    "to_date": "date",
    "bool_and": "all",
    "bool_or": "any",
}


def evaluate_exprs(df: IbisLazyFrame, /, *exprs: IbisExpr) -> list[tuple[str, ir.Value]]:
    native_results: list[tuple[str, ir.Value]] = []
    for expr in exprs:
        native_series_list = expr(df)
        output_names = expr._evaluate_output_names(df)
        if expr._alias_output_names is not None:
            output_names = expr._alias_output_names(output_names)
        if len(output_names) != len(native_series_list):  # pragma: no cover
            msg = f"Internal error: got output names {output_names}, but only got {len(native_series_list)} results"
            raise AssertionError(msg)
        native_results.extend(zip(output_names, native_series_list))
    return native_results


@lru_cache(maxsize=16)
def native_to_narwhals_dtype(ibis_dtype: IbisDataType, version: Version) -> DType:  # noqa: C901, PLR0912
    dtypes = version.dtypes
    if ibis_dtype.is_int64():
        return dtypes.Int64()
    if ibis_dtype.is_int32():
        return dtypes.Int32()
    if ibis_dtype.is_int16():
        return dtypes.Int16()
    if ibis_dtype.is_int8():
        return dtypes.Int8()
    if ibis_dtype.is_uint64():
        return dtypes.UInt64()
    if ibis_dtype.is_uint32():
        return dtypes.UInt32()
    if ibis_dtype.is_uint16():
        return dtypes.UInt16()
    if ibis_dtype.is_uint8():
        return dtypes.UInt8()
    if ibis_dtype.is_boolean():
        return dtypes.Boolean()
    if ibis_dtype.is_float64():
        return dtypes.Float64()
    if ibis_dtype.is_float32():
        return dtypes.Float32()
    if ibis_dtype.is_string():
        return dtypes.String()
    if ibis_dtype.is_date():
        return dtypes.Date()
    if is_timestamp(ibis_dtype):
        _unit = cast("TimestampUnit", ibis_dtype.unit)
        return dtypes.Datetime(time_unit=_unit.value, time_zone=ibis_dtype.timezone)
    if is_interval(ibis_dtype):
        _time_unit = ibis_dtype.unit.value
        if _time_unit not in {"ns", "us", "ms", "s"}:  # pragma: no cover
            msg = f"Unsupported interval unit: {_time_unit}"
            raise NotImplementedError(msg)
        return dtypes.Duration(_time_unit)
    if is_array(ibis_dtype):
        if ibis_dtype.length:
            return dtypes.Array(
                native_to_narwhals_dtype(ibis_dtype.value_type, version),
                ibis_dtype.length,
            )
        return dtypes.List(native_to_narwhals_dtype(ibis_dtype.value_type, version))
    if is_struct(ibis_dtype):
        return dtypes.Struct(
            [
                dtypes.Field(name, native_to_narwhals_dtype(dtype, version))
                for name, dtype in ibis_dtype.items()
            ]
        )
    if is_decimal(ibis_dtype):
        # Same default as in ibis.Decimal.{to_polars, to_pyarrow}
        # https://github.com/ibis-project/ibis/blob/5028d8e5e1921d5de48f59fad48b86c0de541b0d/ibis/formats/polars.py#L82-L86
        # https://github.com/ibis-project/ibis/blob/5028d8e5e1921d5de48f59fad48b86c0de541b0d/ibis/formats/pyarrow.py#L165-L177
        # According to their own comment:
        # > set default precision and scale to something; unclear how to choose this
        # source: (see https://github.com/ibis-project/ibis/blob/5028d8e5e1921d5de48f59fad48b86c0de541b0d/ibis/formats/pyarrow.py#L166)
        scale = 9 if ibis_dtype.scale is None else ibis_dtype.scale
        return dtypes.Decimal(precision=ibis_dtype.precision, scale=scale)
    if ibis_dtype.is_time():
        return dtypes.Time()
    if ibis_dtype.is_binary():
        return dtypes.Binary()
    return dtypes.Unknown()  # pragma: no cover


def is_timestamp(obj: IbisDataType) -> TypeIs[ibis_dtypes.Timestamp]:
    return obj.is_timestamp()


def is_interval(obj: IbisDataType) -> TypeIs[ibis_dtypes.Interval]:
    return obj.is_interval()


def is_array(obj: IbisDataType) -> TypeIs[ibis_dtypes.Array[Any]]:
    return obj.is_array()


def is_struct(obj: IbisDataType) -> TypeIs[ibis_dtypes.Struct]:
    return obj.is_struct()


def is_floating(obj: IbisDataType) -> TypeIs[ibis_dtypes.Floating]:
    return obj.is_floating()


def is_decimal(obj: IbisDataType) -> TypeIs[ibis_dtypes.Decimal]:
    return obj.is_decimal()


dtypes = Version.MAIN.dtypes
NW_TO_IBIS_DTYPES: Mapping[type[DType], IbisDataType] = {
    dtypes.Float64: ibis_dtypes.Float64(),
    dtypes.Float32: ibis_dtypes.Float32(),
    dtypes.Binary: ibis_dtypes.Binary(),
    dtypes.String: ibis_dtypes.String(),
    dtypes.Boolean: ibis_dtypes.Boolean(),
    dtypes.Date: ibis_dtypes.Date(),
    dtypes.Time: ibis_dtypes.Time(),
    dtypes.Int8: ibis_dtypes.Int8(),
    dtypes.Int16: ibis_dtypes.Int16(),
    dtypes.Int32: ibis_dtypes.Int32(),
    dtypes.Int64: ibis_dtypes.Int64(),
    dtypes.UInt8: ibis_dtypes.UInt8(),
    dtypes.UInt16: ibis_dtypes.UInt16(),
    dtypes.UInt32: ibis_dtypes.UInt32(),
    dtypes.UInt64: ibis_dtypes.UInt64(),
}
# Enum support: https://github.com/ibis-project/ibis/issues/10991
UNSUPPORTED_DTYPES = (dtypes.Int128, dtypes.UInt128, dtypes.Categorical, dtypes.Enum)


def narwhals_to_native_dtype(dtype: IntoDType, version: Version) -> IbisDataType:
    dtypes = version.dtypes
    base_type = dtype.base_type()
    if ibis_type := NW_TO_IBIS_DTYPES.get(base_type):
        return ibis_type
    if isinstance_or_issubclass(dtype, dtypes.Datetime):
        return ibis_dtypes.Timestamp.from_unit(dtype.time_unit, timezone=dtype.time_zone)
    if isinstance_or_issubclass(dtype, dtypes.Duration):
        return ibis_dtypes.Interval(unit=dtype.time_unit)  # pyright: ignore[reportArgumentType]
    if isinstance_or_issubclass(dtype, dtypes.List):
        inner = narwhals_to_native_dtype(dtype.inner, version)
        return ibis_dtypes.Array(value_type=inner)
    if isinstance_or_issubclass(dtype, dtypes.Struct):
        fields = [
            (field.name, narwhals_to_native_dtype(field.dtype, version))
            for field in dtype.fields
        ]
        return ibis_dtypes.Struct.from_tuples(fields)
    if isinstance_or_issubclass(dtype, dtypes.Array):
        inner = narwhals_to_native_dtype(dtype.inner, version)
        return ibis_dtypes.Array(value_type=inner, length=dtype.size)
    if isinstance_or_issubclass(dtype, dtypes.Decimal):
        return ibis_dtypes.Decimal(dtype.precision, dtype.scale)
    if issubclass(base_type, UNSUPPORTED_DTYPES):
        msg = f"Converting to {base_type.__name__} dtype is not supported for Ibis."
        raise NotImplementedError(msg)
    msg = f"Unknown dtype: {dtype}"  # pragma: no cover
    raise AssertionError(msg)


def timedelta_to_ibis_interval(td: timedelta) -> ibis.expr.types.temporal.IntervalScalar:
    return ibis.interval(days=td.days, seconds=td.seconds, microseconds=td.microseconds)


def function(name: str, *args: ir.Value | PythonLiteral) -> ir.Value:
    # Workaround SQL vs Ibis differences.
    if name == "row_number":
        return ibis.row_number() + lit(1)
    if name == "least":
        return ibis.least(*args)
    if name == "greatest":
        return ibis.greatest(*args)
    expr = args[0]
    if name == "var_pop":
        return cast("ir.NumericColumn", expr).var(how="pop")
    if name == "var_samp":
        return cast("ir.NumericColumn", expr).var(how="sample")
    if name == "stddev_pop":
        return cast("ir.NumericColumn", expr).std(how="pop")
    if name == "stddev_samp":
        return cast("ir.NumericColumn", expr).std(how="sample")
    if name == "substr":
        # Ibis is 0-indexed here, SQL is 1-indexed
        return cast("ir.StringColumn", expr).substr(args[1] - 1, *args[2:])  # type: ignore[operator]  # pyright: ignore[reportArgumentType]
    return getattr(expr, FUNCTION_REMAPPING.get(name, name))(*args[1:])



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import narwhals as nw
from tests.utils import (
    PANDAS_VERSION,
    PYARROW_VERSION,
    Constructor,
    ConstructorEager,
    assert_equal_data,
)

if TYPE_CHECKING:
    from narwhals.typing import DTypeBackend

data = {"a": [-1, 0, 1, 2, 4]}

expected = [float("nan"), 0, 1, 1.414213, 2]


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_sqrt_expr(constructor: Constructor) -> None:
    df = nw.from_native(constructor(data))
    result = df.select(nw.col("a").sqrt())
    assert_equal_data(result, {"a": expected})


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_sqrt_series(constructor_eager: ConstructorEager) -> None:
    series = nw.from_native(constructor_eager(data), eager_only=True)["a"]
    result = series.sqrt()
    assert_equal_data({"a": result}, {"a": expected})


PYARROW_UNAVAILABLE = PYARROW_VERSION == (0, 0, 0)
reason = "nullable types require pandas2+"
require_pd_2_1 = pytest.mark.skipif(
    PANDAS_VERSION < (2, 1, 0) or PYARROW_UNAVAILABLE, reason=reason
)
require_pd_2_0 = pytest.mark.skipif(PANDAS_VERSION < (2, 0, 0), reason=reason)


@pytest.mark.parametrize(
    "dtype_backend",
    [
        pytest.param("pyarrow", marks=require_pd_2_1),
        pytest.param("numpy_nullable", marks=require_pd_2_0),
        None,
    ],
)
def test_sqrt_dtype_pandas(dtype_backend: DTypeBackend) -> None:
    pytest.importorskip("pandas")
    import pandas as pd

    s = pd.Series([1.0, None, 2.0], name="a", dtype="float32", index=[8, 7, 6])
    if dtype_backend:
        s = s.convert_dtypes(dtype_backend=dtype_backend)

    # if working with nullable types, type is cast float32->Int64 by convert_dtypes
    # then Int64->Float64 by the operation
    expected_dtype = "float32" if not dtype_backend else "float64"
    expected = pd.Series(
        [1.0, None, 1.414213], name="a", dtype=expected_dtype, index=[8, 7, 6]
    )
    if dtype_backend:
        expected = expected.convert_dtypes(dtype_backend=dtype_backend)
    result = nw.from_native(s, series_only=True).sqrt().to_native()
    pd.testing.assert_series_equal(result, expected)


@pytest.mark.skipif(PANDAS_VERSION < (2, 1, 0), reason="nullable types require pandas2+")
def test_sqrt_dtype_pandas_pyarrow() -> None:
    pytest.importorskip("pandas")
    pytest.importorskip("pyarrow")
    import pandas as pd

    s = pd.Series([1.0, None, 2.0], name="a", dtype="Float32[pyarrow]", index=[8, 7, 6])
    result = nw.from_native(s, series_only=True).sqrt().to_native()
    expected = pd.Series(
        [1, None, 1.414213], name="a", dtype="Float64[pyarrow]", index=[8, 7, 6]
    )
    pd.testing.assert_series_equal(result, expected)
